@@ -12,12 +12,12 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: David Reid, dreid@apple.com
 ##
 
 import sys
 
+from twext.python.plistlib import Dict as PlistDict, writePlist
+
 FORMATTERS = {}
 
 def registerFormatter(formatter):
@@ -260,10 +260,26 @@ class PlainFormatter(BaseFormatter):
             self.writeFrequencies(freqs)
             self.write('\n')
 
-        self.write('  User Agents:\n')
+        self.write('  Response Status: total=%d\n' % (len(report['data']['statusStats']),))
+
+        sorted_list = [item for item in report['data']['statusStats'].iteritems()]
+        sorted_list.sort(cmp=lambda x,y: int(x[0]) - int(y[0]))
+        for status, count in sorted_list:
+            self.write('    %s: %s\n' % (status, count))
+        self.write('\n')
+
+        self.write('  User Agents: total=%d\n' % (len(report['data']['userAgents']),))
 
         for ua, count in report['data']['userAgents'].iteritems():
             self.write('    %s: %s\n' % (ua, count))
+        self.write('\n')
+
+        self.write('  Active Users: total=%d\n' % (len(report['data']['activeUsers']),))
+
+        sorted_list = [item for item in report['data']['activeUsers'].iteritems()]
+        sorted_list.sort(cmp=lambda x,y: x[1] - y[1])
+        for user, count in sorted_list:
+            self.write('    %s: %s\n' % (user, count))
 
 registerFormatter(PlainFormatter)
 
@@ -313,25 +329,23 @@ class CsvFormatter(BaseFormatter):
 
 registerFormatter(CsvFormatter)
 
-import plistlib
-
 class PlistFormatter(BaseFormatter):
     name = "plist"
 
     def report_principals(self, report):
-        plist = plistlib.Dict()
+        plist = PlistDict()
 
         plist[report['type']] = list(report['records'])
 
-        plistlib.writePlist(plist, self.dest)
+        writePlist(plist, self.dest)
 
     report_users = report_groups = report_resources = report_locations = report_principals
 
     def report_stats(self, report):
-        plist = plistlib.Dict()
+        plist = PlistDict()
         plist[report['type']] = report['data']
 
-        plistlib.writePlist(plist, self.dest)
+        writePlist(plist, self.dest)
 
     report_logs = report_stats
 
@@ -565,13 +579,35 @@ class HTMLFormatter(BaseFormatter):
             self.writeFrequencies(freqs)
             self.write('<br>\n')
 
-        self.write('<h3>User Agents:</h3>\n')
+        self.write('<h3>Response Status: total=%d</h3>\n' % (len(report['data']['statusStats']),))
+
+        self.write("<table border='1'>")
+        self.write('  <tr><td>Status</td><td># Responses</td></tr>\n')
+
+        sorted_list = [item for item in report['data']['statusStats'].iteritems()]
+        sorted_list.sort(cmp=lambda x,y: int(x[0]) - int(y[0]))
+        for status, count in sorted_list:
+            self.write('  <tr><td>%s:</td><td align=\'right\'>%s</td></tr>\n' % (status, count))
+        self.write('</table>\n')
+
+        self.write('<h3>User Agents: total=%d</h3>\n' % (len(report['data']['userAgents']),))
 
         self.write("<table border='1'>")
         self.write('  <tr><td>User Agent</td><td># Requests</td></tr>\n')
         for ua, count in report['data']['userAgents'].iteritems():
             self.write('  <tr><td>%s:</td><td align=\'right\'>%s</td></tr>\n' % (ua, count))
         self.write('</table>\n')
+
+        self.write('<h3>Active Users: total=%d</h3>\n' % (len(report['data']['activeUsers']),))
+
+        self.write("<table border='1'>")
+        self.write('  <tr><td>User Principal</td><td># Requests</td></tr>\n')
+
+        sorted_list = [item for item in report['data']['activeUsers'].iteritems()]
+        sorted_list.sort(cmp=lambda x,y: y[1] - x[1],)
+        for user, count in sorted_list:
+            self.write('  <tr><td>%s:</td><td align=\'right\'>%s</td></tr>\n' % (user, count))
+        self.write('</table>\n')
         
         self.write('</body>\n</html>\n')
 
