@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2005-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2005-2009 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Wilfredo Sanchez, wsanchez@apple.com
 ##
 
 """
@@ -27,7 +25,9 @@ change.
 See draft spec: http://ietf.webdav.org/caldav/draft-dusseault-caldav.txt
 """
 
-from twisted.python import log
+import datetime
+
+from vobject.icalendar import utc, TimezoneComponent
 
 from twisted.web2.dav import davxml
 
@@ -35,10 +35,9 @@ from twistedcaldav.dateops import clipPeriod, timeRangesOverlap
 from twistedcaldav.ical import Component as iComponent
 from twistedcaldav.ical import Property as iProperty
 from twistedcaldav.ical import parse_date_or_datetime
+from twistedcaldav.log import Logger
 
-from vobject.icalendar import utc
-
-import datetime
+log = Logger()
 
 ##
 # CalDAV objects
@@ -46,9 +45,17 @@ import datetime
 
 caldav_namespace = "urn:ietf:params:xml:ns:caldav"
 
-caldav_compliance = (
+caldav_full_compliance = (
     "calendar-access",
     "calendar-schedule",
+    "calendar-auto-schedule",
+    "calendar-availability",
+    "inbox-availability",
+)
+
+caldav_implicit_compliance = (
+    "calendar-access",
+    "calendar-auto-schedule",
     "calendar-availability",
     "inbox-availability",
 )
@@ -76,15 +83,19 @@ class CalDAVTimeRangeElement (CalDAVEmptyElement):
     CalDAV element containing a time range.
     """
     allowed_attributes = {
-        "start": True,
-        "end"  : True,
+        "start": False,
+        "end"  : False,
     }
 
     def __init__(self, *children, **attributes):
         super(CalDAVTimeRangeElement, self).__init__(*children, **attributes)
 
-        self.start = parse_date_or_datetime(attributes["start"])
-        self.end   = parse_date_or_datetime(attributes["end"  ])
+        # One of start or end must be present
+        if "start" not in attributes and "end" not in attributes:
+            raise ValueError("One of 'start' or 'end' must be present in CALDAV:time-range")
+        
+        self.start = parse_date_or_datetime(attributes["start"]) if "start" in attributes else None
+        self.end = parse_date_or_datetime(attributes["end"]) if "end" in attributes else None
 
 class CalDAVTimeZoneElement (CalDAVTextElement):
     """
@@ -145,7 +156,10 @@ class CalDAVTimeZoneElement (CalDAVTextElement):
         found = False
 
         for subcomponent in calendar.subcomponents():
-            if subcomponent.name() == "VTIMEZONE":
+            if (
+                subcomponent.name() == "VTIMEZONE" and
+                isinstance(subcomponent._vobject, TimezoneComponent)
+            ):
                 if found:
                     return False
                 else:
@@ -273,6 +287,15 @@ class MaxResourceSize (CalDAVTextElement):
     hidden = True
     protected = True
 
+class MaxAttendeesPerInstance (CalDAVTextElement):
+    """
+    Specifies restrictions on a calendar collection.
+    (CalDAV-access-15, section 5.2.9)
+    """
+    name = "max-attendees-per-instance"
+    hidden = True
+    protected = True
+
 class Calendar (CalDAVEmptyElement):
     """
     Denotes a calendar collection.
@@ -373,17 +396,13 @@ class CalendarData (CalDAVElement):
 
     @classmethod
     def fromCalendar(clazz, calendar):
-        assert calendar.name() == "VCALENDAR", "Not a calendar: %r" % (calendar,)
-        return clazz(davxml.PCDATAElement(str(calendar)))
-
-    @classmethod
-    def fromCalendarData(clazz, caldata):
-        """
-        Return a CalendarData element comprised of the supplied calendar data.
-        @param caldata: a string of valid calendar data.
-        @return: a L{CalendarData} element.
-        """
-        return clazz(davxml.PCDATAElement(caldata))
+        if isinstance(calendar, str):
+            return clazz(davxml.PCDATAElement(calendar))
+        elif isinstance(calendar, iComponent):
+            assert calendar.name() == "VCALENDAR", "Not a calendar: %r" % (calendar,)
+            return clazz(davxml.PCDATAElement(str(calendar)))
+        else:
+            raise ValueError("Not a calendar: %s" % (calendar,))
 
     def __init__(self, *children, **attributes):
         super(CalendarData, self).__init__(*children, **attributes)
@@ -468,30 +487,32 @@ class CalendarData (CalDAVElement):
         
         return False
 
-    def elementFromResource(self, resource):
+    def elementFromResource(self, resource, timezone=None):
         """
         Return a new CalendarData element comprised of the possibly filtered
         calendar data from the specified resource. If no filter is being applied
         read the data directly from the resource without parsing it. If a filter
         is required, parse the iCal data and filter using this CalendarData.
         @param resource: the resource whose calendar data is to be returned.
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
         @return: an L{CalendarData} with the (filtered) calendar data.
         """
-        return self.elementFromCalendar(resource.iCalendar())
+        return self.elementFromCalendar(resource.iCalendarText(), timezone)
 
-    def elementFromCalendar(self, calendar):
+    def elementFromCalendar(self, calendar, timezone=None):
         """
         Return a new CalendarData element comprised of the possibly filtered
         calendar.
         @param calendar: the calendar that is to be filtered and returned.
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
         @return: an L{CalendarData} with the (filtered) calendar data.
         """
         
         # Check for filtering or not
-        filtered = self.getFromICalendar(calendar)
+        filtered = self.getFromICalendar(calendar, timezone)
         return CalendarData.fromCalendar(filtered)
 
-    def elementFromResourceWithAccessRestrictions(self, resource, access):
+    def elementFromResourceWithAccessRestrictions(self, resource, access, timezone=None):
         """
         Return a new CalendarData element comprised of the possibly filtered
         calendar data from the specified resource. If no filter is being applied
@@ -502,11 +523,12 @@ class CalendarData (CalDAVElement):
 
         @param resource: the resource whose calendar data is to be returned.
         @param access: private event access restriction level.
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
         @return: an L{CalendarData} with the (filtered) calendar data.
         """
-        return self.elementFromCalendarWithAccessRestrictions(resource.iCalendar(), access)
+        return self.elementFromCalendarWithAccessRestrictions(resource.iCalendarText(), access, timezone)
 
-    def elementFromCalendarWithAccessRestrictions(self, calendar, access):
+    def elementFromCalendarWithAccessRestrictions(self, calendar, access, timezone=None):
         """
         Return a new CalendarData element comprised of the possibly filtered
         calendar.
@@ -515,11 +537,12 @@ class CalendarData (CalDAVElement):
 
         @param calendar: the calendar that is to be filtered and returned.
         @param access: private event access restriction level.
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
         @return: an L{CalendarData} with the (filtered) calendar data.
         """
         
         # Do normal filtering first
-        filtered_calendar = self.getFromICalendar(calendar)
+        filtered_calendar = self.getFromICalendar(calendar, timezone)
         
         if access in (iComponent.ACCESS_CONFIDENTIAL, iComponent.ACCESS_RESTRICTED):
             # Create a CALDAV:calendar-data element with the appropriate iCalendar Component/Property
@@ -535,7 +558,7 @@ class CalendarData (CalDAVElement):
             filter = CalendarData(
                 CalendarComponent(
                     
-                    # VCALENDAR proeprties
+                    # VCALENDAR properties
                     Property(name="PRODID"),
                     Property(name="VERSION"),
                     Property(name="CALSCALE"),
@@ -619,37 +642,47 @@ class CalendarData (CalDAVElement):
             )
 
             # Now "filter" the resource calendar data through the CALDAV:calendar-data element
-            return filter.elementFromCalendar(filtered_calendar)
+            return filter.elementFromCalendar(filtered_calendar, timezone)
         else:
             return CalendarData.fromCalendar(filtered_calendar)
 
-    def getFromICalendar(self, calendar):
+    def getFromICalendar(self, calendar, timezone=None):
         """
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
+
         Returns a calendar object containing the data in the given calendar
         which is specified by this CalendarData.
         """
-        if calendar.name() != "VCALENDAR":
+        if calendar is None or isinstance(calendar, str) and not calendar:
             raise ValueError("Not a calendar: %r" % (calendar,))
 
         # Empty element: get all data
         if not self.children: return calendar
 
-        # CalDAV:comp is required 
-        if self.component is None:
-            raise ValueError("CalDAV:calendar-data %s has no CalDAV:comp child" % (self,))
+        # If we were passed a string, parse it out as a Component
+        if isinstance(calendar, str):
+            try:
+                calendar = iComponent.fromString(calendar)
+            except ValueError:
+                raise ValueError("Not a calendar: %r" % (calendar,))
+
+        if calendar is None or calendar.name() != "VCALENDAR":
+            raise ValueError("Not a calendar: %r" % (calendar,))
 
         # Pre-process the calendar data based on expand and limit options
         if self.freebusy_set:
             calendar = self.limitFreeBusy(calendar)
 
-        calendar = self.component.getFromICalendar(calendar)
+        # Filter data based on any provided CALDAV:comp element, or use all current data
+        if self.component is not None:
+            calendar = self.component.getFromICalendar(calendar)
         
         # Post-process the calendar data based on the expand and limit options
         if self.recurrence_set:
             if isinstance(self.recurrence_set, LimitRecurrenceSet):
                 calendar = self.limitRecurrence(calendar)
             elif isinstance(self.recurrence_set, Expand):
-                calendar = self.expandRecurrence(calendar)
+                calendar = self.expandRecurrence(calendar, timezone)
         
         return calendar
 
@@ -676,13 +709,14 @@ class CalendarData (CalDAVElement):
 
         return str(data)
 
-    def expandRecurrence(self, calendar):
+    def expandRecurrence(self, calendar, timezone=None):
         """
         Expand the recurrence set into individual items.
         @param calendar: the L{Component} for the calendar to operate on.
+        @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
         @return: the L{Component} for the result.
         """
-        return calendar.expand(self.recurrence_set.start, self.recurrence_set.end)
+        return calendar.expand(self.recurrence_set.start, self.recurrence_set.end, timezone)
     
     def limitRecurrence(self, calendar):
         """
@@ -797,6 +831,13 @@ class CalendarComponent (CalDAVElement):
         result = iComponent(self.type)
 
         xml_components = self.components
+        xml_properties = self.properties
+
+        # Empty element means do all properties and components
+        if xml_components is None and xml_properties is None:
+            xml_components = AllComponents()
+            xml_properties = AllProperties()
+
         if xml_components is not None:
             if xml_components == AllComponents():
                 for ical_subcomponent in component.subcomponents():
@@ -807,7 +848,6 @@ class CalendarComponent (CalDAVElement):
                         if ical_subcomponent.name() == xml_subcomponent.type:
                             result.addComponent(xml_subcomponent.getFromICalendar(ical_subcomponent))
 
-        xml_properties = self.properties
         if xml_properties is not None:
             if xml_properties == AllProperties():
                 for ical_property in component.properties():
@@ -908,9 +948,17 @@ class Filter (CalDAVElement):
         # We need to prepare ourselves for a time-range query by pre-calculating
         # the set of instances up to the latest time-range limit. That way we can
         # avoid having to do some form of recurrence expansion for each query sub-part.
-        maxend = self.children[0].getLastExpandTime()
+        maxend, isStartTime = self.getmaxtimerange()
         if maxend:
-            instances = component.expandTimeRanges(maxend)
+            if isStartTime:
+                if component.isRecurringUnbounded():
+                    # Unbounded recurrence is always within a start-only time-range
+                    instances = None
+                else:
+                    # Expand the instances up to infinity
+                    instances = component.expandTimeRanges(datetime.datetime(2100, 1, 1, 0, 0, 0, tzinfo=utc), ignoreInvalidInstances=True)
+            else:
+                instances = component.expandTimeRanges(maxend, ignoreInvalidInstances=True)
         else:
             instances = None
         self.children[0].setInstances(instances)
@@ -952,6 +1000,13 @@ class Filter (CalDAVElement):
         self.children[0].settzinfo(utc)
         return utc
 
+    def getmaxtimerange(self):
+        """
+        Get the date farthest into the future in any time-range elements
+        """
+        
+        return self.children[0].getmaxtimerange(None, False)
+
 class ComponentFilter (CalDAVFilterElement):
     """
     Limits a search to only the chosen component types.
@@ -972,7 +1027,7 @@ class ComponentFilter (CalDAVFilterElement):
         """
         Returns True if the given calendar item (which is a component)
         matches this filter, False otherwise.
-        This specialzation uses the instance matching option of the time-range filter
+        This specialization uses the instance matching option of the time-range filter
         to minimize instance expansion.
         """
 
@@ -1008,28 +1063,6 @@ class ComponentFilter (CalDAVFilterElement):
             return not self.defined
         return self.defined
         
-    def getLastExpandTime(self):
-        """
-        Get the latest time-range end value from any time-range element in this
-        or child comp-filter elements.
-        @return: the L{datetime.datetime} corrsponding to the max. time to expand to,
-                    or None if there is no time-range
-        """
-        
-        # Look for time-range in this filter
-        if self.qualifier and self.defined:
-            maxend = self.qualifier.end
-        else:
-            maxend = None
-            
-        # Now look at each comp-filter element in this one
-        for compfilter in [x for x in self.filters if isinstance(x, ComponentFilter)]:
-            end = compfilter.getLastExpandTime()
-            if end and ((maxend is None) or (end > maxend)):
-                maxend = end
-
-        return maxend
-    
     def setInstances(self, instances):
         """
         Give the list of instances to each comp-filter element.
@@ -1057,7 +1090,7 @@ class ComponentFilter (CalDAVFilterElement):
                 log.msg("Top-level comp-filter must be VCALENDAR, instead: %s" % (self.filter_name,))
                 return False
         elif level == 1:
-            # Dissallow VCALENDAR, VALARM, STANDARD, DAYLIGHT, AVAILABLE at the top, everything else is OK
+            # Disallow VCALENDAR, VALARM, STANDARD, DAYLIGHT, AVAILABLE at the top, everything else is OK
             if self.filter_name in ("VCALENDAR", "VALARM", "STANDARD", "DAYLIGHT", "AVAILABLE"):
                 log.msg("comp-filter wrong component type: %s" % (self.filter_name,))
                 return False
@@ -1067,7 +1100,7 @@ class ComponentFilter (CalDAVFilterElement):
                 log.msg("time-range cannot be used with component %s" % (self.filter_name,))
                 return False
         elif level == 2:
-            # Dissallow VCALENDAR, VTIMEZONE, VEVENT, VTODO, VJOURNAL, VFREEBUSY, VAVAILABILITY at the top, everything else is OK
+            # Disallow VCALENDAR, VTIMEZONE, VEVENT, VTODO, VJOURNAL, VFREEBUSY, VAVAILABILITY at the top, everything else is OK
             if (self.filter_name in ("VCALENDAR", "VTIMEZONE", "VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VAVAILABILITY")):
                 log.msg("comp-filter wrong sub-component type: %s" % (self.filter_name,))
                 return False
@@ -1077,7 +1110,7 @@ class ComponentFilter (CalDAVFilterElement):
                 log.msg("time-range cannot be used with sub-component %s" % (self.filter_name,))
                 return False
         else:
-            # Dissallow all std iCal components anywhere else
+            # Disallow all standard iCal components anywhere else
             if (self.filter_name in ("VCALENDAR", "VTIMEZONE", "VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VALARM", "STANDARD", "DAYLIGHT", "AVAILABLE")) or timerange:
                 log.msg("comp-filter wrong standard component type: %s" % (self.filter_name,))
                 return False
@@ -1113,6 +1146,29 @@ class ComponentFilter (CalDAVFilterElement):
         for x in self.filters:
             x.settzinfo(tzinfo)
 
+    def getmaxtimerange(self, currentMaximum, currentIsStartTime):
+        """
+        Get the date furthest into the future in any time-range elements
+        
+        @param currentMaximum: current future value to compare with
+        @type currentMaximum: L{datetime.datetime}
+        """
+        
+        # Give tzinfo to any TimeRange we have
+        isStartTime = False
+        if isinstance(self.qualifier, TimeRange):
+            isStartTime = self.qualifier.end is None
+            compareWith = self.qualifier.start if isStartTime else self.qualifier.end
+            if currentMaximum is None or currentMaximum < compareWith:
+                currentMaximum = compareWith
+                currentIsStartTime = isStartTime
+        
+        # Pass down to sub components/properties
+        for x in self.filters:
+            currentMaximum, currentIsStartTime = x.getmaxtimerange(currentMaximum, currentIsStartTime)
+
+        return currentMaximum, currentIsStartTime
+
 class PropertyFilter (CalDAVFilterElement):
     """
     Limits a search to specific properties.
@@ -1183,6 +1239,25 @@ class PropertyFilter (CalDAVFilterElement):
         if isinstance(self.qualifier, TimeRange):
             self.qualifier.settzinfo(tzinfo)
 
+    def getmaxtimerange(self, currentMaximum, currentIsStartTime):
+        """
+        Get the date furthest into the future in any time-range elements
+        
+        @param currentMaximum: current future value to compare with
+        @type currentMaximum: L{datetime.datetime}
+        """
+        
+        # Give tzinfo to any TimeRange we have
+        isStartTime = False
+        if isinstance(self.qualifier, TimeRange):
+            isStartTime = self.qualifier.end is None
+            compareWith = self.qualifier.start if isStartTime else self.qualifier.end
+            if currentMaximum is None or currentMaximum < compareWith:
+                currentMaximum = compareWith
+                currentIsStartTime = isStartTime
+
+        return currentMaximum, currentIsStartTime
+
 class ParameterFilter (CalDAVFilterElement):
     """
     Limits a search to specific parameters.
@@ -1316,8 +1391,8 @@ class TextMatch (CalDAVTextElement):
             return False, False
 
         for value in values:
-            # NB Its possible that we have a text list value which appears as a Pythin list,
-            # so we need to check for that an iterate over thr list.
+            # NB Its possible that we have a text list value which appears as a Python list,
+            # so we need to check for that an iterate over the list.
             if isinstance(value, list):
                 for subvalue in value:
                     matched, result = _textCompare(subvalue)
@@ -1364,17 +1439,17 @@ class TimeRange (CalDAVTimeRangeElement):
         @return:      True if valid, False otherwise
         """
         
-        if not isinstance(self.start, datetime.datetime):
+        if self.start is not None and not isinstance(self.start, datetime.datetime):
             log.msg("start attribute in <time-range> is not a date-time: %s" % (self.start,))
             return False
-        if not isinstance(self.end, datetime.datetime):
+        if self.end is not None and not isinstance(self.end, datetime.datetime):
             log.msg("end attribute in <time-range> is not a date-time: %s" % (self.end,))
             return False
-        if self.start.tzinfo != utc:
+        if self.start is not None and self.start.tzinfo != utc:
             log.msg("start attribute in <time-range> is not UTC: %s" % (self.start,))
             return False
-        if self.end.tzinfo != utc:
-            log.msg("end attribute in <time-range> is not UTC: %s" % (self.start,))
+        if self.end is not None and self.end.tzinfo != utc:
+            log.msg("end attribute in <time-range> is not UTC: %s" % (self.end,))
             return False
 
         # No other tests
@@ -1400,7 +1475,19 @@ class TimeRange (CalDAVTimeRangeElement):
         if component is None:
             return False
         
-        assert instances is not None, "Failure to expand instance for time-range filter: %r" % (self,)
+        assert instances is not None or self.end is None, "Failure to expand instance for time-range filter: %r" % (self,)
+        
+        # Special case open-ended unbounded
+        if instances is None:
+            if component.getRecurrenceIDUTC() is None:
+                return True
+            else:
+                # See if the overridden component's start is past the start
+                start, _ignore_end = component.getEffectiveStartEnd()
+                if start is None:
+                    return True
+                else:
+                    return start >= self.start
 
         # Handle alarms as a special case
         alarms = (component.name() == "VALARM")
@@ -1553,14 +1640,46 @@ class CalendarUserAddressSet (CalDAVElement):
 class CalendarFreeBusySet (CalDAVElement):
     """
     The list of calendar URIs that contribute to free-busy for this principal's calendar user.
-    (CalDAV-schedule, section x.x.x)
+    This was defined in the old caldav scheduling spec but has been removed from the new one.
+    We still need to support it for backwards compatibility.
     """
     name = "calendar-free-busy-set"
     hidden = True
 
     allowed_children = { (davxml.dav_namespace, "href"): (0, None) }
 
-class ScheduleInboxURL (CalDAVTextElement):
+class ScheduleCalendarTransp (CalDAVElement):
+    """
+    Indicates whether a calendar should be used for freebusy lookups.
+    """
+    name = "schedule-calendar-transp"
+
+    allowed_children = {
+        (caldav_namespace,     "opaque"      ): (0, 1),
+        (caldav_namespace,     "transparent" ): (0, 1),
+    }
+
+class Opaque (CalDAVEmptyElement):
+    """
+    Indicates that a calendar is used in freebusy lookups.
+    """
+    name = "opaque"
+
+class Transparent (CalDAVEmptyElement):
+    """
+    Indicates that a calendar is not used in freebusy lookups.
+    """
+    name = "transparent"
+
+class ScheduleDefaultCalendarURL (CalDAVElement):
+    """
+    A single href indicating which calendar is the default for scheduling.
+    """
+    name = "schedule-default-calendar-URL"
+
+    allowed_children = { (davxml.dav_namespace, "href"): (0, 1) }
+
+class ScheduleInboxURL (CalDAVElement):
     """
     A principal property to indicate the schedule INBOX for the principal.
     (CalDAV-schedule, section x.x.x)
@@ -1571,7 +1690,7 @@ class ScheduleInboxURL (CalDAVTextElement):
 
     allowed_children = { (davxml.dav_namespace, "href"): (0, 1) }
 
-class ScheduleOutboxURL (CalDAVTextElement):
+class ScheduleOutboxURL (CalDAVElement):
     """
     A principal property to indicate the schedule OUTBOX for the principal.
     (CalDAV-schedule, section x.x.x)
@@ -1596,11 +1715,11 @@ class Originator (CalDAVElement):
 
 class Recipient (CalDAVElement):
     """
-    A property on resources in schedule Inbox indicating the Recipients targetted
+    A property on resources in schedule Inbox indicating the Recipients targeted
     by the SCHEDULE operation.
     (CalDAV-schedule, section x.x.x)
     
-    The recipient for whom this reponse is for.
+    The recipient for whom this response is for.
     (CalDAV-schedule, section x.x.x)
     """
     name = "recipient"
@@ -1609,6 +1728,15 @@ class Recipient (CalDAVElement):
 
     allowed_children = { (davxml.dav_namespace, "href"): (0, None) } # NB Minimum is zero because this is a property name
 
+class ScheduleTag (CalDAVTextElement):
+    """
+    Property on scheduling resources.
+    (CalDAV-schedule, section x.x.x)
+    """
+    name = "schedule-tag"
+    hidden = True
+    protected = True
+
 class ScheduleInbox (CalDAVEmptyElement):
     """
     Denotes the resource type of a calendar schedule Inbox.
@@ -1618,7 +1746,7 @@ class ScheduleInbox (CalDAVEmptyElement):
 
 class ScheduleOutbox (CalDAVEmptyElement):
     """
-    Denotes the resourcetype of a calendar schedule Outbox.
+    Denotes the resource type of a calendar schedule Outbox.
     (CalDAV-schedule-xx, section x.x.x)
     """
     name = "schedule-outbox"
@@ -1661,6 +1789,27 @@ class Schedule (CalDAVEmptyElement):
     """
     name = "schedule"
     
+class ScheduleDeliver (CalDAVEmptyElement):
+    """
+    Privilege which controls scheduling messages going into the Inbox.
+    (CalDAV-schedule, section x.x.x)
+    """
+    name = "schedule-deliver"
+    
+class ScheduleSend (CalDAVEmptyElement):
+    """
+    Privilege which controls the ability to send scheduling messages.
+    (CalDAV-schedule, section x.x.x)
+    """
+    name = "schedule-send"
+    
+class CalendarUserType (CalDAVTextElement):
+    """
+    The CALDAV:calendar-user-type property from section 9.2.4 of caldav-sched-05
+    """
+    name = "calendar-user-type"
+    protected = True
+
 ##
 # Extensions to davxml.ResourceType
 ##
