@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2006-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2006-2009 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Wilfredo Sanchez, wsanchez@apple.com
 ##
 
 """
@@ -22,34 +20,48 @@ Implements a directory-backed calendar hierarchy.
 
 __all__ = [
     "uidsResourceName",
+   #"DirectoryCalendarProvisioningResource",
     "DirectoryCalendarHomeProvisioningResource",
     "DirectoryCalendarHomeTypeProvisioningResource",
     "DirectoryCalendarHomeUIDProvisioningResource",
     "DirectoryCalendarHomeResource",
 ]
 
-from twisted.internet.defer import succeed
+from twisted.internet.defer import succeed, inlineCallbacks, returnValue
 from twisted.web2 import responsecode
 from twisted.web2.dav import davxml
 from twisted.web2.http import HTTPError
 from twisted.web2.dav.util import joinURL
-from twisted.web2.dav.resource import TwistedACLInheritable, TwistedQuotaRootProperty
-from twisted.web2.http import HTTPError
+from twisted.web2.dav.resource import TwistedACLInheritable
 
 from twistedcaldav import caldavxml
 from twistedcaldav.config import config
 from twistedcaldav.dropbox import DropBoxHomeResource
 from twistedcaldav.extensions import ReadOnlyResourceMixIn, DAVResource
-from twistedcaldav.notifications import NotificationsCollectionResource
+from twistedcaldav.freebusyurl import FreeBusyURLResource
 from twistedcaldav.resource import CalDAVResource
 from twistedcaldav.schedule import ScheduleInboxResource, ScheduleOutboxResource
 from twistedcaldav.directory.idirectory import IDirectoryService
+from twistedcaldav.directory.wiki import getWikiACL
 from twistedcaldav.directory.resource import AutoProvisioningResourceMixIn
 
+from twistedcaldav.log import Logger
+log = Logger()
+
 # Use __underbars__ convention to avoid conflicts with directory resource types.
 uidsResourceName = "__uids__"
 
-class DirectoryCalendarHomeProvisioningResource (AutoProvisioningResourceMixIn, ReadOnlyResourceMixIn, DAVResource):
+
+class DirectoryCalendarProvisioningResource (
+    AutoProvisioningResourceMixIn,
+    ReadOnlyResourceMixIn,
+    DAVResource,
+):
+    def defaultAccessControlList(self):
+        return config.ProvisioningResourceACL
+
+
+class DirectoryCalendarHomeProvisioningResource (DirectoryCalendarProvisioningResource):
     """
     Resource which provisions calendar home collections as needed.    
     """
@@ -107,7 +119,7 @@ class DirectoryCalendarHomeProvisioningResource (AutoProvisioningResourceMixIn,
         if uidResource is None:
             return None
         else:
-            return uidResource.getChild(record.guid)
+            return uidResource.getChild(record.uid)
 
     ##
     # DAV
@@ -116,14 +128,8 @@ class DirectoryCalendarHomeProvisioningResource (AutoProvisioningResourceMixIn,
     def isCollection(self):
         return True
 
-    ##
-    # ACL
-    ##
 
-    def defaultAccessControlList(self):
-        return readOnlyACL
-
-class DirectoryCalendarHomeTypeProvisioningResource (AutoProvisioningResourceMixIn, ReadOnlyResourceMixIn, DAVResource):
+class DirectoryCalendarHomeTypeProvisioningResource (DirectoryCalendarProvisioningResource):
     """
     Resource which provisions calendar home collections of a specific
     record type as needed.
@@ -158,11 +164,18 @@ class DirectoryCalendarHomeTypeProvisioningResource (AutoProvisioningResourceMix
         return self._parent.homeForDirectoryRecord(record)
 
     def listChildren(self):
-        return (
-            record.shortName
-            for record in self.directory.listRecords(self.recordType)
-            if record.enabledForCalendaring
-        )
+        if config.EnablePrincipalListings:
+
+            def _recordShortnameExpand():
+                for record in self.directory.listRecords(self.recordType):
+                    if record.enabledForCalendaring:
+                        for shortName in record.shortNames:
+                            yield shortName
+
+            return _recordShortnameExpand()
+        else:
+            # Not a listable collection
+            raise HTTPError(responsecode.FORBIDDEN)
 
     def createSimilarFile(self, path):
         raise HTTPError(responsecode.NOT_FOUND)
@@ -178,9 +191,6 @@ class DirectoryCalendarHomeTypeProvisioningResource (AutoProvisioningResourceMix
     # ACL
     ##
 
-    def defaultAccessControlList(self):
-        return readOnlyACL
-
     def principalCollections(self):
         return self._parent.principalCollections()
 
@@ -188,7 +198,7 @@ class DirectoryCalendarHomeTypeProvisioningResource (AutoProvisioningResourceMix
         return self._parent.principalForRecord(record)
 
 
-class DirectoryCalendarHomeUIDProvisioningResource (AutoProvisioningResourceMixIn, ReadOnlyResourceMixIn, DAVResource):
+class DirectoryCalendarHomeUIDProvisioningResource (DirectoryCalendarProvisioningResource):
     def __init__(self, parent):
         """
         @param parent: the parent of this resource
@@ -209,7 +219,7 @@ class DirectoryCalendarHomeUIDProvisioningResource (AutoProvisioningResourceMixI
             return self
 
         if record is None:
-            record = self.directory.recordWithGUID(name)
+            record = self.directory.recordWithUID(name)
             if record is None:
                 return None
 
@@ -230,9 +240,6 @@ class DirectoryCalendarHomeUIDProvisioningResource (AutoProvisioningResourceMixI
     # ACL
     ##
 
-    def defaultAccessControlList(self):
-        return readOnlyACL
-
     def principalCollections(self):
         return self.parent.principalCollections()
 
@@ -265,58 +272,72 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
             childlist += (
                 ("dropbox", DropBoxHomeResource),
             )
-        if config.EnableNotifications:
+        if config.FreeBusyURL.Enabled:
             childlist += (
-                ("notifications", NotificationsCollectionResource),
+                ("freebusy", FreeBusyURLResource),
             )
         for name, cls in childlist:
             child = self.provisionChild(name)
             assert isinstance(child, cls), "Child %r is not a %s: %r" % (name, cls.__name__, child)
             self.putChild(name, child)
 
-    def provision(self):
-        # If an ACL property does not currently exist, create one from
-        # the defaultACL
-        if not self.hasDeadProperty(davxml.ACL):
-            self.writeDeadProperty(self.defaultAccessControlList())
-        
-        super(DirectoryCalendarHomeResource, self).provision()
-
     def provisionDefaultCalendars(self):
-        self.provision()
 
-        childName = "calendar"
-        childURL = joinURL(self.url(), childName)
-        child = self.provisionChild(childName)
-        assert isinstance(child, CalDAVResource), "Child %r is not a %s: %r" % (childName, CalDAVResource.__name__, child)
+        # Disable notifications during provisioning
+        if hasattr(self, "clientNotifier"):
+            self.clientNotifier.disableNotify()
+
+        def setupFreeBusy(_):
+            # Default calendar is initially opaque to freebusy
+            child.writeDeadProperty(caldavxml.ScheduleCalendarTransp(caldavxml.Opaque()))
 
-        def setupChild(_):
-            # Set calendar-free-busy-set on inbox
-            inbox = self.getChild("inbox")
             # FIXME: Shouldn't have to call provision() on another resource
             # We cheat here because while inbox will auto-provision itself when located,
             # we need to write a dead property to it pre-emptively.
-            # Possible fix: store the free/busy set property on this resource instead.
+            # This will go away once we remove the free-busy-set property on inbox.
+
+            # Set calendar-free-busy-set on inbox
+            inbox = self.getChild("inbox")
             inbox.provision()
-            inbox.writeDeadProperty(caldavxml.CalendarFreeBusySet(davxml.HRef(childURL)))
+            inbox.processFreeBusyCalendar(childURL, True)
+
+            # Default calendar is marked as the default for scheduling
+            inbox.writeDeadProperty(caldavxml.ScheduleDefaultCalendarURL(davxml.HRef(childURL)))
 
             return self
 
-        d = child.createCalendarCollection()
-        d.addCallback(setupChild)
+        try:
+            self.provision()
+
+            childName = "calendar"
+            childURL = joinURL(self.url(), childName)
+            child = self.provisionChild(childName)
+            assert isinstance(child, CalDAVResource), "Child %r is not a %s: %r" % (childName, CalDAVResource.__name__, child)
+
+            d = child.createCalendarCollection()
+            d.addCallback(setupFreeBusy)
+        except:
+            # We want to make sure to re-enable notifications, so do so
+            # if there is an immediate exception above, or via errback, below
+            if hasattr(self, "clientNotifier"):
+                self.clientNotifier.enableNotify(None)
+            raise
+
+        # Re-enable notifications
+        if hasattr(self, "clientNotifier"):
+            d.addCallback(self.clientNotifier.enableNotify)
+            d.addErrback(self.clientNotifier.enableNotify)
+
         return d
 
     def provisionChild(self, name):
         raise NotImplementedError("Subclass must implement provisionChild()")
 
     def url(self):
-        return joinURL(self.parent.url(), self.record.guid)
-        ##
-        ## While the underlying primary location is GUID-based, we want
-        ## the canonical user-facing location to be recordType &
-        ## shortName-based, because that's friendlier.
-        ##
-        #return joinURL(self.parent.parent.getChild(self.record.recordType).url(), self.record.shortName)
+        return joinURL(self.parent.url(), self.record.uid)
+
+    def canonicalURL(self, request):
+        return succeed(self.url())
 
     ##
     # DAV
@@ -325,6 +346,9 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
     def isCollection(self):
         return True
 
+    def http_COPY(self, request):
+        return responsecode.FORBIDDEN
+
     ##
     # ACL
     ##
@@ -332,15 +356,13 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
     def owner(self, request):
         return succeed(davxml.HRef(self.principalForRecord().principalURL()))
 
+    def ownerPrincipal(self, request):
+        return succeed(self.principalForRecord())
+
     def defaultAccessControlList(self):
         myPrincipal = self.principalForRecord()
 
         aces = (
-            # DAV:read access for authenticated users.
-            davxml.ACE(
-                davxml.Principal(davxml.Authenticated()),
-                davxml.Grant(davxml.Privilege(davxml.Read())),
-            ),
             # Inheritable DAV:all access for the resource's associated principal.
             davxml.ACE(
                 davxml.Principal(davxml.HRef(myPrincipal.principalURL())),
@@ -355,20 +377,33 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
                 TwistedACLInheritable(),
             ),
         )
+
+        # Give read access to config.ReadPrincipals
+        aces += config.ReadACEs
+
+        # Give all access to config.AdminPrincipals
+        aces += config.AdminACEs
         
         if config.EnableProxyPrincipals:
             aces += (
-                # DAV:read access for this principal's calendar-proxy-read users.
+                # DAV:read/DAV:read-current-user-privilege-set access for this principal's calendar-proxy-read users.
                 davxml.ACE(
                     davxml.Principal(davxml.HRef(joinURL(myPrincipal.principalURL(), "calendar-proxy-read/"))),
-                    davxml.Grant(davxml.Privilege(davxml.Read())),
+                    davxml.Grant(
+                        davxml.Privilege(davxml.Read()),
+                        davxml.Privilege(davxml.ReadCurrentUserPrivilegeSet()),
+                    ),
                     davxml.Protected(),
                     TwistedACLInheritable(),
                 ),
-                # DAV:read/DAV:write access for this principal's calendar-proxy-write users.
+                # DAV:read/DAV:read-current-user-privilege-set/DAV:write access for this principal's calendar-proxy-write users.
                 davxml.ACE(
                     davxml.Principal(davxml.HRef(joinURL(myPrincipal.principalURL(), "calendar-proxy-write/"))),
-                    davxml.Grant(davxml.Privilege(davxml.Read()), davxml.Privilege(davxml.Write())),
+                    davxml.Grant(
+                        davxml.Privilege(davxml.Read()),
+                        davxml.Privilege(davxml.ReadCurrentUserPrivilegeSet()),
+                        davxml.Privilege(davxml.Write()),
+                    ),
                     davxml.Protected(),
                     TwistedACLInheritable(),
                 ),
@@ -376,6 +411,19 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
 
         return davxml.ACL(*aces)
 
+    @inlineCallbacks
+    def accessControlList(self, request, inheritance=True, expanding=False, inherited_aces=None):
+
+        wikiACL = (yield getWikiACL(self, request))
+        if wikiACL is not None:
+            # ACL depends on wiki server...
+            log.debug("Wiki ACL: %s" % (wikiACL.toxml(),))
+            returnValue(wikiACL)
+        else:
+            # ...otherwise permissions are fixed, and are not subject to
+            # inheritance rules, etc.
+            returnValue(self.defaultAccessControlList())
+
     def principalCollections(self):
         return self.parent.principalCollections()
 
@@ -388,29 +436,17 @@ class DirectoryCalendarHomeResource (AutoProvisioningResourceMixIn, CalDAVResour
 
     def hasQuotaRoot(self, request):
         """
+        Always get quota root value from config.
+
         @return: a C{True} if this resource has quota root, C{False} otherwise.
         """
-        return self.hasDeadProperty(TwistedQuotaRootProperty) or config.UserQuota
+        return config.UserQuota != 0
     
     def quotaRoot(self, request):
         """
+        Always get quota root value from config.
+
         @return: a C{int} containing the maximum allowed bytes if this collection
             is quota-controlled, or C{None} if not quota controlled.
         """
-        if self.hasDeadProperty(TwistedQuotaRootProperty):
-            return int(str(self.readDeadProperty(TwistedQuotaRootProperty)))
-        else:
-            return config.UserQuota
-
-##
-# Utilities
-##
-
-# DAV:read access for authenticated users.
-readOnlyACL = davxml.ACL(
-    davxml.ACE(
-        davxml.Principal(davxml.Authenticated()),
-        davxml.Grant(davxml.Privilege(davxml.Read())),
-        davxml.Protected(),
-    ),
-)
+        return config.UserQuota if config.UserQuota != 0 else None
