@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2005-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2005-2009 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,26 +12,19 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Wilfredo Sanchez, wsanchez@apple.com
 ##
 
-#from twisted.web2 import responsecode
-#from twisted.web2.iweb import IResponse
-#from twisted.web2.dav import davxml
-#from twisted.web2.dav.util import davXMLFromStream
-#from twisted.web2.test.test_server import SimpleRequest
-#from twistedcaldav import caldavxml
-
 import os
 
+from twisted.cred.credentials import UsernamePassword
+from twisted.internet.defer import inlineCallbacks
 from twisted.web2.dav import davxml
 from twisted.web2.dav.fileop import rmdir
 from twisted.web2.dav.resource import AccessDeniedError
 from twisted.web2.test.test_server import SimpleRequest
-from twisted.web2.dav.test.util import serialize
 
 from twistedcaldav.static import CalendarHomeProvisioningFile
+from twistedcaldav.config import config
 from twistedcaldav.directory.apache import BasicDirectoryService, DigestDirectoryService
 from twistedcaldav.directory.directory import DirectoryService
 from twistedcaldav.directory.test.test_apache import basicUserFile, digestUserFile, groupFile, digestRealm
@@ -42,13 +35,10 @@ from twistedcaldav.directory.principal import DirectoryPrincipalTypeProvisioning
 from twistedcaldav.directory.principal import DirectoryPrincipalResource
 from twistedcaldav.directory.principal import DirectoryCalendarPrincipalResource
 
+from twistedcaldav.cache import DisabledCacheNotifier
+
 import twistedcaldav.test.util
 
-directoryServices = (
-    BasicDirectoryService(digestRealm, basicUserFile, groupFile),
-    DigestDirectoryService(digestRealm, digestUserFile, groupFile),
-    XMLDirectoryService(xmlFile),
-)
 
 class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
     """
@@ -56,19 +46,36 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
     """
     def setUp(self):
         super(ProvisionedPrincipals, self).setUp()
-        
+
+        self.directoryServices = (
+            BasicDirectoryService(
+                {
+                    'realmName' : digestRealm,
+                    'userFile' : basicUserFile,
+                    'groupFile' : groupFile,
+                }
+            ),
+            DigestDirectoryService(
+                {
+                    'realmName' : digestRealm,
+                    'userFile' : digestUserFile,
+                    'groupFile' : groupFile,
+                }
+            ),
+            XMLDirectoryService(
+                {
+                    'xmlFile' : xmlFile,
+                }
+            ),
+        )
+
         # Set up a principals hierarchy for each service we're testing with
         self.principalRootResources = {}
-        for directory in directoryServices:
+        for directory in self.directoryServices:
             name = directory.__class__.__name__
             url = "/" + name + "/"
-            path = os.path.join(self.docroot, url[1:])
 
-            if os.path.exists(path):
-                rmdir(path)
-            os.mkdir(path)
-
-            provisioningResource = DirectoryPrincipalProvisioningResource(path, url, directory)
+            provisioningResource = DirectoryPrincipalProvisioningResource(url, directory)
 
             self.site.resource.putChild(name, provisioningResource)
 
@@ -88,7 +95,7 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
 
         DirectoryPrincipalResource.principalURL(),
         """
-        for directory in directoryServices:
+        for directory in self.directoryServices:
             #print "\n -> %s" % (directory.__class__.__name__,)
             provisioningResource = self.principalRootResources[directory.__class__.__name__]
 
@@ -113,8 +120,8 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
                 self.assertEquals(set((provisioningURL,)), set(pc.principalCollectionURL() for pc in principalCollections))
 
                 shortNames = set(typeResource.listChildren())
-                self.assertEquals(shortNames, set(r.shortName for r in directory.listRecords(recordType)))
-                
+                self.assertEquals(shortNames, set(r.shortNames[0] for r in directory.listRecords(recordType)))
+
                 for shortName in shortNames:
                     #print "     -> %s" % (shortName,)
                     recordResource = typeResource.getChild(shortName)
@@ -142,7 +149,7 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
         DirectoryPrincipalProvisioningResource.principalForShortName()
         """
         for provisioningResource, recordType, recordResource, record in self._allRecords():
-            principal = provisioningResource.principalForShortName(recordType, record.shortName)
+            principal = provisioningResource.principalForShortName(recordType, record.shortNames[0])
             self.failIf(principal is None)
             self.assertEquals(record, principal.record)
 
@@ -150,20 +157,33 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
         """
         DirectoryPrincipalProvisioningResource.principalForUser()
         """
-        for directory in directoryServices:
+        for directory in self.directoryServices:
             provisioningResource = self.principalRootResources[directory.__class__.__name__]
 
             for user in directory.listRecords(DirectoryService.recordType_users):
-                userResource = provisioningResource.principalForUser(user.shortName)
+                userResource = provisioningResource.principalForUser(user.shortNames[0])
                 self.failIf(userResource is None)
                 self.assertEquals(user, userResource.record)
 
-    def test_principalForGUID(self):
+    def test_principalForAuthID(self):
         """
-        DirectoryPrincipalProvisioningResource.principalForGUID()
+        DirectoryPrincipalProvisioningResource.principalForAuthID()
+        """
+        for directory in self.directoryServices:
+            provisioningResource = self.principalRootResources[directory.__class__.__name__]
+
+            for user in directory.listRecords(DirectoryService.recordType_users):
+                creds = UsernamePassword(user.shortNames[0], "bogus")
+                userResource = provisioningResource.principalForAuthID(creds)
+                self.failIf(userResource is None)
+                self.assertEquals(user, userResource.record)
+
+    def test_principalForUID(self):
+        """
+        DirectoryPrincipalProvisioningResource.principalForUID()
         """
         for provisioningResource, recordType, recordResource, record in self._allRecords():
-            principal = provisioningResource.principalForGUID(record.guid)
+            principal = provisioningResource.principalForUID(record.uid)
             self.failIf(principal is None)
             self.assertEquals(record, principal.record)
 
@@ -195,19 +215,11 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
                 else:
                     self.failIf(principal is not None)
 
-    def test_autoSchedule(self):
-        """
-        DirectoryPrincipalProvisioningResource.principalForCalendarUserAddress()
-        """
-        for provisioningResource, recordType, recordResource, record in self._allRecords():
-            principal = provisioningResource.principalForRecord(record)
-            self.failIf(principal is None)
-            if record.enabledForCalendaring:
-                self.assertEquals(record.autoSchedule, principal.autoSchedule())
-                if record.shortName == "gemini":
-                    self.assertTrue(principal.autoSchedule())
-                else:
-                    self.assertFalse(principal.autoSchedule())
+        # Explicitly check the disabled record
+        self.failIf(provisioningResource.principalForCalendarUserAddress("mailto:nocalendar@example.com") is not None)
+        self.failIf(provisioningResource.principalForCalendarUserAddress("urn:uuid:543D28BA-F74F-4D5F-9243-B3E3A61171E5") is not None)
+        self.failIf(provisioningResource.principalForCalendarUserAddress("/principals/users/nocalendar/") is not None)
+        self.failIf(provisioningResource.principalForCalendarUserAddress("/principals/__uids__/543D28BA-F74F-4D5F-9243-B3E3A61171E5/") is not None)
 
     def test_enabledForCalendaring(self):
         """
@@ -222,12 +234,52 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
                 self.assertTrue(isinstance(principal, DirectoryPrincipalResource))
                 self.assertFalse(isinstance(principal, DirectoryCalendarPrincipalResource))
 
+    def test_enabledAsOrganizer(self):
+        """
+        DirectoryPrincipalProvisioningResource.principalForCalendarUserAddress()
+        """
+        
+        ok_types = (
+            DirectoryService.recordType_users,
+        )
+        for provisioningResource, recordType, recordResource, record in self._allRecords():
+            
+            if record.enabledForCalendaring:
+                principal = provisioningResource.principalForRecord(record)
+                self.failIf(principal is None)
+                self.assertEqual(principal.enabledAsOrganizer(), recordType in ok_types)
+
+        config.Scheduling.Options.AllowGroupAsOrganizer = True
+        config.Scheduling.Options.AllowLocationAsOrganizer = True
+        config.Scheduling.Options.AllowResourceAsOrganizer = True
+        ok_types = (
+            DirectoryService.recordType_users,
+            DirectoryService.recordType_groups,
+            DirectoryService.recordType_locations,
+            DirectoryService.recordType_resources,
+        )
+        for provisioningResource, recordType, recordResource, record in self._allRecords():
+            
+            if record.enabledForCalendaring:
+                principal = provisioningResource.principalForRecord(record)
+                self.failIf(principal is None)
+                self.assertEqual(principal.enabledAsOrganizer(), recordType in ok_types)
+
     # FIXME: Run DirectoryPrincipalProvisioningResource tests on DirectoryPrincipalTypeProvisioningResource also
 
     ##
     # DirectoryPrincipalResource
     ##
 
+    def test_cacheNotifier(self):
+        """
+        Each DirectoryPrincipalResource should have a cacheNotifier attribute
+        that is an instance of XattrCacheChangeNotifier
+        """
+        for provisioningResource, recordType, recordResource, record in self._allRecords():
+            self.failUnless(isinstance(recordResource.cacheNotifier,
+                                       DisabledCacheNotifier))
+
     def test_displayName(self):
         """
         DirectoryPrincipalResource.displayName()
@@ -235,28 +287,23 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
         for provisioningResource, recordType, recordResource, record in self._allRecords():
             self.failUnless(recordResource.displayName())
 
+    @inlineCallbacks
     def test_groupMembers(self):
         """
         DirectoryPrincipalResource.groupMembers()
         """
         for provisioningResource, recordType, recordResource, record in self._allRecords():
-            self.failUnless(set(record.members()).issubset(set(r.record for r in recordResource.groupMembers())))
+            members = yield recordResource.groupMembers()
+            self.failUnless(set(record.members()).issubset(set(r.record for r in members)))
 
+    @inlineCallbacks
     def test_groupMemberships(self):
         """
         DirectoryPrincipalResource.groupMemberships()
         """
         for provisioningResource, recordType, recordResource, record in self._allRecords():
-            self.failUnless(set(record.groups()).issubset(set(r.record for r in recordResource.groupMemberships() if hasattr(r, "record"))))
-
-    def test_proxies(self):
-        """
-        DirectoryPrincipalResource.proxies()
-        """
-        for provisioningResource, recordType, recordResource, record in self._allRecords():
-            if record.enabledForCalendaring:
-                self.failUnless(set(record.proxies()).issubset(set(r.record for r in recordResource.proxies())))
-                self.assertEqual(record.hasEditableProxyMembership(), recordResource.hasEditableProxyMembership())
+            memberships = yield recordResource.groupMemberships()
+            self.failUnless(set(record.groups()).issubset(set(r.record for r in memberships if hasattr(r, "record"))))
 
     def test_principalUID(self):
         """
@@ -294,7 +341,7 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
         # Need to create a calendar home provisioner for each service.
         calendarRootResources = {}
 
-        for directory in directoryServices:
+        for directory in self.directoryServices:
             url = "/homes_" + directory.__class__.__name__ + "/"
             path = os.path.join(self.docroot, url[1:])
 
@@ -305,68 +352,100 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
             provisioningResource = CalendarHomeProvisioningFile(path, directory, url)
 
             calendarRootResources[directory.__class__.__name__] = provisioningResource
-        
+
         # Calendar home provisioners should result in calendar homes.
         for provisioningResource, recordType, recordResource, record in self._allRecords():
             if record.enabledForCalendaring:
                 homeURLs = tuple(recordResource.calendarHomeURLs())
                 self.failUnless(homeURLs)
-    
+
                 calendarRootURL = calendarRootResources[record.service.__class__.__name__].url()
-    
+
                 inboxURL = recordResource.scheduleInboxURL()
                 outboxURL = recordResource.scheduleOutboxURL()
-    
+
                 self.failUnless(inboxURL)
                 self.failUnless(outboxURL)
-    
+
                 for homeURL in homeURLs:
                     self.failUnless(homeURL.startswith(calendarRootURL))
-    
+
                     if inboxURL and inboxURL.startswith(homeURL):
                         self.failUnless(len(inboxURL) > len(homeURL))
                         self.failUnless(inboxURL.endswith("/"))
                         inboxURL = None
-    
+
                     if outboxURL and outboxURL.startswith(homeURL):
                         self.failUnless(len(outboxURL) > len(homeURL))
                         self.failUnless(outboxURL.endswith("/"))
                         outboxURL = None
-    
+
                 self.failIf(inboxURL)
                 self.failIf(outboxURL)
 
+    @inlineCallbacks
     def test_defaultAccessControlList_principals(self):
         """
         Default access controls for principals.
         """
-        def work():
-            for provisioningResource, recordType, recordResource, record in self._allRecords():
-                for args in _authReadOnlyPrivileges(recordResource, recordResource.principalURL()):
-                    yield args
-
-        return serialize(self._checkPrivileges, work())
+        for provisioningResource, recordType, recordResource, record in self._allRecords():
+            for args in _authReadOnlyPrivileges(self, recordResource, recordResource.principalURL()):
+                yield self._checkPrivileges(*args)
 
+    @inlineCallbacks
     def test_defaultAccessControlList_provisioners(self):
         """
         Default access controls for principal provisioning resources.
         """
-        def work():
-            for directory in directoryServices:
-                #print "\n -> %s" % (directory.__class__.__name__,)
-                provisioningResource = self.principalRootResources[directory.__class__.__name__]
+        for directory in self.directoryServices:
+            #print "\n -> %s" % (directory.__class__.__name__,)
+            provisioningResource = self.principalRootResources[directory.__class__.__name__]
 
-                for args in _authReadOnlyPrivileges(provisioningResource, provisioningResource.principalCollectionURL()):
-                    yield args
+            for args in _authReadOnlyPrivileges(self, provisioningResource, provisioningResource.principalCollectionURL()):
+                yield self._checkPrivileges(*args)
 
-                for recordType in provisioningResource.listChildren():
-                    #print "   -> %s" % (recordType,)
-                    typeResource = provisioningResource.getChild(recordType)
+            for recordType in provisioningResource.listChildren():
+                #print "   -> %s" % (recordType,)
+                typeResource = provisioningResource.getChild(recordType)
 
-                    for args in _authReadOnlyPrivileges(typeResource, typeResource.principalCollectionURL()):
-                        yield args
+                for args in _authReadOnlyPrivileges(self, typeResource, typeResource.principalCollectionURL()):
+                    yield self._checkPrivileges(*args)
 
-        return serialize(self._checkPrivileges, work())
+    def test_propertyToField(self):
+
+        class stubElement(object):
+            def __init__(self, ns, name):
+                self.ns = ns
+                self.name = name
+
+            def qname(self):
+                return self.ns, self.name
+
+        provisioningResource = self.principalRootResources['BasicDirectoryService']
+
+        expected = (
+            ("DAV:", "displayname", "morgen", "fullName", "morgen"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-type", "INDIVIDUAL", "recordType", "users"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-type", "GROUP", "recordType", "groups"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-type", "RESOURCE", "recordType", "resources"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-type", "ROOM", "recordType", "locations"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "/principals/__uids__/AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA/", "guid", "AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "http://example.com:8008/principals/__uids__/AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA/", "guid", "AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "urn:uuid:AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA", "guid", "AAAAAAAA-AAAA-AAAA-AAAA-AAAAAAAAAAAA"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "/principals/users/example/", "recordName", "example"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "https://example.com:8443/principals/users/example/", "recordName", "example"),
+            ("urn:ietf:params:xml:ns:caldav", "calendar-user-address-set", "mailto:example@example.com", "emailAddresses", "example@example.com"),
+            ("http://calendarserver.org/ns/", "first-name", "morgen", "firstName", "morgen"),
+            ("http://calendarserver.org/ns/", "last-name", "sagen", "lastName", "sagen"),
+            ("http://calendarserver.org/ns/", "email-address-set", "example@example.com", "emailAddresses", "example@example.com"),
+        )
+
+        for ns, property, match, field, converted in expected:
+            el = stubElement(ns, property)
+            self.assertEquals(
+                provisioningResource.propertyToField(el, match),
+                (field, converted)
+            )
 
     def _allRecords(self):
         """
@@ -378,7 +457,7 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
             C{record} is the directory service record
             for each record in each directory in C{directoryServices}.
         """
-        for directory in directoryServices:
+        for directory in self.directoryServices:
             provisioningResource = self.principalRootResources[directory.__class__.__name__]
             for recordType in directory.recordTypes():
                 for record in directory.listRecords(recordType):
@@ -410,13 +489,17 @@ class ProvisionedPrincipals (twistedcaldav.test.util.TestCase):
         d.addCallback(gotResource)
         return d
 
-def _authReadOnlyPrivileges(resource, url):
-    for principal, privilege, allowed in (
-        ( davxml.All()             , davxml.Read()  , False ),
-        ( davxml.All()             , davxml.Write() , False ),
-        ( davxml.Unauthenticated() , davxml.Read()  , False ),
-        ( davxml.Unauthenticated() , davxml.Write() , False ),
-        ( davxml.Authenticated()   , davxml.Read()  , True  ),
-        ( davxml.Authenticated()   , davxml.Write() , False ),
-    ):
+def _authReadOnlyPrivileges(self, resource, url):
+    items = []
+    for provisioningResource, recordType, recordResource, record in self._allRecords():
+        if recordResource == resource:
+            items.append(( davxml.HRef().fromString(recordResource.principalURL()), davxml.Read()  , True ))
+            items.append(( davxml.HRef().fromString(recordResource.principalURL()), davxml.Write() , True ))
+        else:
+            items.append(( davxml.HRef().fromString(recordResource.principalURL()), davxml.Read()  , True ))
+            items.append(( davxml.HRef().fromString(recordResource.principalURL()), davxml.Write() , False ))
+    items.append(( davxml.Unauthenticated() , davxml.Read()  , False ))
+    items.append(( davxml.Unauthenticated() , davxml.Write() , False ))
+            
+    for principal, privilege, allowed in items:
         yield resource, url, principal, privilege, allowed
