@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2005-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2005-2009 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Wilfredo Sanchez, wsanchez@apple.com
 ##
 
 """
@@ -24,45 +22,54 @@ change.
 """
 
 __all__ = [
+    "db_basename",
+    "ReservationError",
+    "MemcachedUIDReserver",
     "Index",
     "IndexSchedule",
+    "IndexedSearchException",
 ]
 
 import datetime
 import os
 import time
+import hashlib
 
 try:
     import sqlite3 as sqlite
 except ImportError:
     from pysqlite2 import dbapi2 as sqlite
 
-from twisted.python import log
+from vobject.icalendar import utc
+
+from twisted.internet.defer import maybeDeferred, succeed
 
 from twistedcaldav.ical import Component
 from twistedcaldav.query import calendarquery
 from twistedcaldav.sql import AbstractSQLDatabase
 from twistedcaldav.sql import db_prefix
 from twistedcaldav import caldavxml
+from twistedcaldav.instance import InvalidOverriddenInstanceError
+from twistedcaldav.log import Logger, LoggingMixIn
+from twistedcaldav.config import config
+from twistedcaldav.memcachepool import CachePoolUserMixIn
 
-from vobject.icalendar import utc
+log = Logger()
 
 db_basename = db_prefix + "sqlite"
 schema_version = "6"
 collection_types = {"Calendar": "Regular Calendar Collection", "iTIP": "iTIP Calendar Collection"}
 
-reservation_timeout_secs = 5 * 60
-
 #
-# Duration into the future through which recurrances are expanded in the index
+# Duration into the future through which recurrences are expanded in the index
 # by default.  This is a caching parameter which affects the size of the index;
 # it does not affect search results beyond this period, but it may affect
 # performance of such a search.
 #
-default_future_expansion_duration = datetime.timedelta(days=356*1)
+default_future_expansion_duration = datetime.timedelta(days=365*1)
 
 #
-# Maximum duration into the future through which recurrances are expanded in the
+# Maximum duration into the future through which recurrences are expanded in the
 # index.  This is a caching parameter which affects the size of the index; it
 # does not affect search results beyond this period, but it may affect
 # performance of such a search.
@@ -75,15 +82,18 @@ default_future_expansion_duration = datetime.timedelta(days=356*1)
 # period.  Searches beyond this period will always be relatively expensive for
 # resources with occurances beyond this period.
 #
-maximum_future_expansion_duration = datetime.timedelta(days=356*5)
+maximum_future_expansion_duration = datetime.timedelta(days=365*5)
 
 class ReservationError(LookupError):
     """
-    Attempt to reserve a UID which is already reserved or to unreverse a UID
+    Attempt to reserve a UID which is already reserved or to unreserve a UID
     which is not reserved.
     """
 
-class AbstractCalendarIndex(AbstractSQLDatabase):
+class IndexedSearchException(ValueError):
+    pass
+
+class AbstractCalendarIndex(AbstractSQLDatabase, LoggingMixIn):
     """
     Calendar collection index abstract base class that defines the apis for the index.
     This will be subclassed for the two types of index behaviour we need: one for
@@ -98,7 +108,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         """
         self.resource = resource
         db_filename = os.path.join(self.resource.fp.path, db_basename)
-        super(AbstractCalendarIndex, self).__init__(db_filename)
+        super(AbstractCalendarIndex, self).__init__(db_filename, False)
 
     def create(self):
         """
@@ -180,7 +190,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         for name in self.resourceNamesForUID(uid):
             assert result is None, "More than one resource with UID %s in calendar collection %r" % (uid, self)
             result = name
-            
+
         return result
 
     def resourceUIDForName(self, name):
@@ -194,7 +204,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
 
         return uid
 
-    def addResource(self, name, calendar, fast=False):
+    def addResource(self, name, calendar, fast=False, reCreate=False):
         """
         Adding or updating an existing resource.
         To check for an update we attempt to get an existing UID
@@ -208,7 +218,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         oldUID = self.resourceUIDForName(name)
         if oldUID is not None:
             self._delete_from_db(name, oldUID)
-        self._add_to_db(name, calendar)
+        self._add_to_db(name, calendar, reCreate=reCreate)
         if not fast:
             self._db_commit()
 
@@ -222,7 +232,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         if uid is not None:
             self._delete_from_db(name, uid)
             self._db_commit()
-    
+
     def resourceExists(self, name):
         """
         Determines whether the specified resource name exists in the index.
@@ -231,7 +241,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         """
         uid = self._db_value_for_sql("select UID from RESOURCE where NAME = :1", name)
         return uid is not None
-    
+
     def resourcesExist(self, names):
         """
         Determines whether the specified resource name exists in the index.
@@ -246,16 +256,17 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         statement += ")"
         results = self._db_values_for_sql(statement, *names)
         return results
-    
-    def searchValid(self, filter):
-        if isinstance(filter, caldavxml.Filter):
-            qualifiers = calendarquery.sqlcalendarquery(filter)
-        else:
-            qualifiers = None
-            
-        return qualifiers is not None
 
-    def search(self, filter):
+
+    def testAndUpdateIndex(self, minDate):
+        # Find out if the index is expanded far enough
+        names = self.notExpandedBeyond(minDate)
+        # Actually expand recurrence max
+        for name in names:
+            self.log_info("Search falls outside range of index for %s %s" % (name, minDate))
+            self.reExpandResource(name, minDate)
+
+    def indexedSearch(self, filter):
         """
         Finds resources matching the given qualifiers.
         @param filter: the L{Filter} for the calendar-query to execute.
@@ -264,19 +275,37 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
             C{name} is the resource name, C{uid} is the resource UID, and
             C{type} is the resource iCalendar component type.x
         """
-        # FIXME: Don't forget to use maximum_future_expansion_duration when we
-        # start caching...
-        
-        # Make sure we have a proper Filter element and get the partial SQL statement to use.
+
+        # Make sure we have a proper Filter element and get the partial SQL
+        # statement to use.
         if isinstance(filter, caldavxml.Filter):
             qualifiers = calendarquery.sqlcalendarquery(filter)
+            if qualifiers is not None:
+                # Determine how far we need to extend the current expansion of
+                # events. If we have an open-ended time-range we will expand one
+                # year past the start. That should catch bounded recurrences - unbounded
+                # will have been indexed with an "infinite" value always included.
+                maxDate, isStartDate = filter.getmaxtimerange()
+                if maxDate:
+                    maxDate = maxDate.date()
+                    if isStartDate:
+                        maxDate += datetime.timedelta(days=365)
+                    self.testAndUpdateIndex(maxDate)
+            else:
+                # We cannot handler this filter in an indexed search
+                raise IndexedSearchException()
+
         else:
             qualifiers = None
-        if qualifiers is not None:
-            rowiter = self._db_execute("select DISTINCT RESOURCE.NAME, RESOURCE.UID, RESOURCE.TYPE" + qualifiers[0], *qualifiers[1])
-        else:
+
+        # Perform the search
+        if qualifiers is None:
             rowiter = self._db_execute("select NAME, UID, TYPE from RESOURCE")
-            
+        else:
+            rowiter = self._db_execute("select DISTINCT RESOURCE.NAME, RESOURCE.UID, RESOURCE.TYPE" + qualifiers[0], *qualifiers[1])
+
+        # Check result for missing resources
+
         for row in rowiter:
             name = row[0]
             if self.resource.getChild(name.encode("utf-8")):
@@ -286,13 +315,33 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
                         % (name, self.resource))
                 self.deleteResource(name)
 
+    def bruteForceSearch(self):
+        """
+        List the whole index and tests for existence, updating the index
+        @return: all resources in the index
+        """
+        # List all resources
+        rowiter = self._db_execute("select NAME, UID, TYPE from RESOURCE")
+
+        # Check result for missing resources:
+
+        for row in rowiter:
+            name = row[0]
+            if self.resource.getChild(name.encode("utf-8")):
+                yield row
+            else:
+                log.err("Calendar resource %s is missing from %s. Removing from index."
+                        % (name, self.resource))
+                self.deleteResource(name)
+
+
     def _db_version(self):
         """
         @return: the schema version assigned to this index.
         """
         return schema_version
-        
-    def _add_to_db(self, name, calendar, cursor = None):
+
+    def _add_to_db(self, name, calendar, cursor = None, expand_until=None, reCreate=False):
         """
         Records the given calendar resource in the index with the given name.
         Resource names and UIDs must both be unique; only one resource name may
@@ -304,7 +353,7 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
             contents.
         """
         raise NotImplementedError
-    
+
     def _delete_from_db(self, name, uid):
         """
         Deletes the specified entry from all dbs.
@@ -312,12 +361,12 @@ class AbstractCalendarIndex(AbstractSQLDatabase):
         @param uid: the uid of the resource to delete.
         """
         raise NotImplementedError
-    
+
 class CalendarIndex (AbstractCalendarIndex):
     """
     Calendar index - abstract class for indexer that indexes calendar objects in a collection.
     """
-    
+
     def __init__(self, resource):
         """
         @param resource: the L{twistedcaldav.static.CalDAVFile} resource to
@@ -335,7 +384,7 @@ class CalendarIndex (AbstractCalendarIndex):
         #   NAME: Last URI component (eg. <uid>.ics, RESOURCE primary key)
         #   UID: iCalendar UID (may or may not be unique)
         #   TYPE: iCalendar component type
-        #   RECURRANCE_MAX: Highest date of recurrance expansion
+        #   RECURRANCE_MAX: Highest date of recurrence expansion
         #
         if uidunique:
             q.execute(
@@ -393,7 +442,23 @@ class CalendarIndex (AbstractCalendarIndex):
                 """
             )
 
-    def _add_to_db(self, name, calendar, cursor = None):
+    def notExpandedBeyond(self, minDate):
+        """
+        Gives all resources which have not been expanded beyond a given date
+        in the index
+        """
+        return self._db_values_for_sql("select NAME from RESOURCE where RECURRANCE_MAX < :1", minDate)
+
+    def reExpandResource(self, name, expand_until):
+        """
+        Given a resource name, remove it from the database and re-add it
+        with a longer expansion.
+        """
+        calendar = self.resource.getChild(name).iCalendar()
+        self._add_to_db(name, calendar, expand_until=expand_until, reCreate=True)
+        self._db_commit()
+
+    def _add_to_db(self, name, calendar, cursor = None, expand_until=None, reCreate=False):
         """
         Records the given calendar resource in the index with the given name.
         Resource names and UIDs must both be unique; only one resource name may
@@ -406,14 +471,35 @@ class CalendarIndex (AbstractCalendarIndex):
         """
         uid = calendar.resourceUID()
 
-        expand_max = datetime.date.today() + default_future_expansion_duration
+        # Decide how far to expand based on the component
+        master = calendar.masterComponent()
+        if master is None or not calendar.isRecurring() and not calendar.isRecurringUnbounded():
+            # When there is no master we have a set of overridden components - index them all.
+            # When there is one instance - index it.
+            # When bounded - index all.
+            expand = datetime.datetime(2100, 1, 1, 0, 0, 0, tzinfo=utc)
+        else:
+            if expand_until:
+                expand = expand_until
+            else:
+                expand = datetime.date.today() + default_future_expansion_duration
+    
+            if expand > (datetime.date.today() + maximum_future_expansion_duration):
+                raise IndexedSearchException
+
+        try:
+            instances = calendar.expandTimeRanges(expand, ignoreInvalidInstances=reCreate)
+        except InvalidOverriddenInstanceError, e:
+            log.err("Invalid instance %s when indexing %s in %s" % (e.rid, name, self.resource,))
+            raise
+
+        self._delete_from_db(name, uid)
 
-        instances = calendar.expandTimeRanges(expand_max)
         for key in instances:
             instance = instances[key]
             start = instance.start.replace(tzinfo=utc)
             end = instance.end.replace(tzinfo=utc)
-            float = ('N', 'Y')[instance.start.tzinfo is None]
+            float = 'Y' if instance.start.tzinfo is None else 'N'
             self._db_execute(
                 """
                 insert into TIMESPAN (NAME, FLOAT, START, END)
@@ -421,13 +507,26 @@ class CalendarIndex (AbstractCalendarIndex):
                 """, name, float, start, end
             )
 
+        # Special - for unbounded recurrence we insert a value for "infinity"
+        # that will allow an open-ended time-range to always match it.
+        if calendar.isRecurringUnbounded():
+            start = datetime.datetime(2100, 1, 1, 0, 0, 0, tzinfo=utc)
+            end = datetime.datetime(2100, 1, 1, 1, 0, 0, tzinfo=utc)
+            float = 'N'
+            self._db_execute(
+                """
+                insert into TIMESPAN (NAME, FLOAT, START, END)
+                values (:1, :2, :3, :4)
+                """, name, float, start, end
+            )
+             
         self._db_execute(
             """
             insert into RESOURCE (NAME, UID, TYPE, RECURRANCE_MAX)
             values (:1, :2, :3, :4)
             """, name, uid, calendar.resourceType(), instances.limit
         )
-    
+
     def _delete_from_db(self, name, uid):
         """
         Deletes the specified entry from all dbs.
@@ -436,12 +535,170 @@ class CalendarIndex (AbstractCalendarIndex):
         """
         self._db_execute("delete from TIMESPAN where NAME = :1", name)
         self._db_execute("delete from RESOURCE where NAME = :1", name)
-    
+
+
+def wrapInDeferred(f):
+    def _(*args, **kwargs):
+        return maybeDeferred(f, *args, **kwargs)
+
+    return _
+
+
+class MemcachedUIDReserver(CachePoolUserMixIn, LoggingMixIn):
+    def __init__(self, index, cachePool=None):
+        self.index = index
+        self._cachePool = cachePool
+
+    def _key(self, uid):
+        return 'reservation:%s' % (
+            hashlib.md5('%s:%s' % (uid,
+                                   self.index.resource.fp.path)).hexdigest())
+
+    def reserveUID(self, uid):
+        uid = uid.encode('utf-8')
+        self.log_debug("Reserving UID %r @ %r" % (
+                uid,
+                self.index.resource.fp.path))
+
+        def _handleFalse(result):
+            if result is False:
+                raise ReservationError(
+                    "UID %s already reserved for calendar collection %s."
+                    % (uid, self.index.resource)
+                    )
+
+        d = self.getCachePool().add(self._key(uid),
+                                    'reserved',
+                                    expireTime=config.UIDReservationTimeOut)
+        d.addCallback(_handleFalse)
+        return d
+
+
+    def unreserveUID(self, uid):
+        uid = uid.encode('utf-8')
+        self.log_debug("Unreserving UID %r @ %r" % (
+                uid,
+                self.index.resource.fp.path))
+
+        def _handleFalse(result):
+            if result is False:
+                raise ReservationError(
+                    "UID %s is not reserved for calendar collection %s."
+                    % (uid, self.index.resource)
+                    )
+
+        d =self.getCachePool().delete(self._key(uid))
+        d.addCallback(_handleFalse)
+        return d
+
+
+    def isReservedUID(self, uid):
+        uid = uid.encode('utf-8')
+        self.log_debug("Is reserved UID %r @ %r" % (
+                uid,
+                self.index.resource.fp.path))
+
+        def _checkValue((flags, value)):
+            if value is None:
+                return False
+            else:
+                return True
+
+        d = self.getCachePool().get(self._key(uid))
+        d.addCallback(_checkValue)
+        return d
+
+
+
+class SQLUIDReserver(object):
+    def __init__(self, index):
+        self.index = index
+
+    @wrapInDeferred
+    def reserveUID(self, uid):
+        """
+        Reserve a UID for this index's resource.
+        @param uid: the UID to reserve
+        @raise ReservationError: if C{uid} is already reserved
+        """
+
+        try:
+            self.index._db_execute("insert into RESERVED (UID, TIME) values (:1, :2)", uid, datetime.datetime.now())
+            self.index._db_commit()
+        except sqlite.IntegrityError:
+            self.index._db_rollback()
+            raise ReservationError(
+                "UID %s already reserved for calendar collection %s."
+                % (uid, self.index.resource)
+            )
+        except sqlite.Error, e:
+            log.err("Unable to reserve UID: %s", (e,))
+            self.index._db_rollback()
+            raise
+
+    def unreserveUID(self, uid):
+        """
+        Unreserve a UID for this index's resource.
+        @param uid: the UID to reserve
+        @raise ReservationError: if C{uid} is not reserved
+        """
+
+        def _cb(result):
+            if result == False:
+                raise ReservationError(
+                    "UID %s is not reserved for calendar collection %s."
+                    % (uid, self.index.resource)
+                    )
+            else:
+                try:
+                    self.index._db_execute(
+                        "delete from RESERVED where UID = :1", uid)
+                    self.index._db_commit()
+                except sqlite.Error, e:
+                    log.err("Unable to unreserve UID: %s", (e,))
+                    self.index._db_rollback()
+                    raise
+
+        d = self.isReservedUID(uid)
+        d.addCallback(_cb)
+        return d
+
+
+    @wrapInDeferred
+    def isReservedUID(self, uid):
+        """
+        Check to see whether a UID is reserved.
+        @param uid: the UID to check
+        @return: True if C{uid} is reserved, False otherwise.
+        """
+
+        rowiter = self.index._db_execute("select UID, TIME from RESERVED where UID = :1", uid)
+        for uid, attime in rowiter:
+            # Double check that the time is within a reasonable period of now
+            # otherwise we probably have a stale reservation
+            tm = time.strptime(attime[:19], "%Y-%m-%d %H:%M:%S")
+            dt = datetime.datetime(year=tm.tm_year, month=tm.tm_mon, day=tm.tm_mday, hour=tm.tm_hour, minute=tm.tm_min, second = tm.tm_sec)
+            if datetime.datetime.now() - dt > datetime.timedelta(seconds=config.UIDReservationTimeOut):
+                try:
+                    self.index._db_execute("delete from RESERVED where UID = :1", uid)
+                    self.index._db_commit()
+                except sqlite.Error, e:
+                    log.err("Unable to unreserve UID: %s", (e,))
+                    self.index._db_rollback()
+                    raise
+                return False
+            else:
+                return True
+
+        return False
+
+
+
 class Index (CalendarIndex):
     """
     Calendar collection index - regular collection that enforces CalDAV UID uniqueness requirement.
     """
-    
+
     def __init__(self, resource):
         """
         @param resource: the L{twistedcaldav.static.CalDAVFile} resource to
@@ -451,80 +708,29 @@ class Index (CalendarIndex):
         assert resource.isCalendarCollection(), "non-calendar collection resource %s has no index." % (resource,)
         super(Index, self).__init__(resource)
 
+        if config.Memcached['ClientEnabled']:
+            self.reserver = MemcachedUIDReserver(self)
+
+        else:
+            self.reserver = SQLUIDReserver(self)
+
     #
     # A dict of sets. The dict keys are calendar collection paths,
     # and the sets contains reserved UIDs for each path.
     #
-    
+
     def reserveUID(self, uid):
-        """
-        Reserve a UID for this index's resource.
-        @param uid: the UID to reserve
-        @raise ReservationError: if C{uid} is already reserved
-        """
+        return self.reserver.reserveUID(uid)
+
 
-        try:
-            self._db_execute("insert into RESERVED (UID, TIME) values (:1, :2)", uid, datetime.datetime.now())
-            self._db_commit()
-        except sqlite.IntegrityError:
-            self._db_rollback()
-            raise ReservationError(
-                "UID %s already reserved for calendar collection %s."
-                % (uid, self.resource)
-            )
-        except sqlite.Error, e:
-            log.err("Unable to reserve UID: %s", (e,))
-            self._db_rollback()
-            raise
-    
     def unreserveUID(self, uid):
-        """
-        Unreserve a UID for this index's resource.
-        @param uid: the UID to reserve
-        @raise ReservationError: if C{uid} is not reserved
-        """
-        
-        if not self.isReservedUID(uid):
-            raise ReservationError(
-                "UID %s is not reserved for calendar collection %s."
-                % (uid, self.resource)
-            )
+        return self.reserver.unreserveUID(uid)
+
 
-        try:
-            self._db_execute("delete from RESERVED where UID = :1", uid)
-            self._db_commit()
-        except sqlite.Error, e:
-            log.err("Unable to unreserve UID: %s", (e,))
-            self._db_rollback()
-            raise
-    
     def isReservedUID(self, uid):
-        """
-        Check to see whether a UID is reserved.
-        @param uid: the UID to check
-        @return: True if C{uid} is reserved, False otherwise.
-        """
-        
-        rowiter = self._db_execute("select UID, TIME from RESERVED where UID = :1", uid)
-        for uid, attime in rowiter:
-            # Double check that the time is within a reasonable period of now
-            # otherwise we probably have a stale reservation
-            tm = time.strptime(attime[:19], "%Y-%m-%d %H:%M:%S")
-            dt = datetime.datetime(year=tm.tm_year, month=tm.tm_mon, day=tm.tm_mday, hour=tm.tm_hour, minute=tm.tm_min, second = tm.tm_sec)
-            if datetime.datetime.now() - dt > datetime.timedelta(seconds=reservation_timeout_secs):
-                try:
-                    self._db_execute("delete from RESERVED where UID = :1", uid)
-                    self._db_commit()
-                except sqlite.Error, e:
-                    log.err("Unable to unreserve UID: %s", (e,))
-                    self._db_rollback()
-                    raise
-                return False
-            else:
-                return True
+        return self.reserver.isReservedUID(uid)
+
 
-        return False
-        
     def isAllowedUID(self, uid, *names):
         """
         Checks to see whether to allow an operation which would add the
@@ -538,27 +744,27 @@ class Index (CalendarIndex):
         """
         rname = self.resourceNameForUID(uid)
         return (rname is None or rname in names)
- 
+
     def _db_type(self):
         """
         @return: the collection type assigned to this index.
         """
         return collection_types["Calendar"]
-        
+
     def _db_init_data_tables(self, q):
         """
         Initialise the underlying database tables.
         @param q:           a database cursor to use.
         """
-        
+
         # Create database where the RESOURCE table has unique UID column.
         self._db_init_data_tables_base(q, True)
 
-    def _db_recreate(self):
+    def _db_recreate(self, do_commit=True):
         """
         Re-create the database tables from existing calendar data.
         """
-        
+
         #
         # Populate the DB with data from already existing resources.
         # This allows for index recovery if the DB file gets
@@ -575,21 +781,21 @@ class Index (CalendarIndex):
                 log.err("Unable to open resource %s: %s" % (name, e))
                 continue
 
+            # FIXME: This is blocking I/O
             try:
-                # FIXME: This is blocking I/O
-                try:
-                    calendar = Component.fromStream(stream)
-                    calendar.validateForCalDAV()
-                except ValueError:
-                    log.err("Non-calendar resource: %s" % (name,))
-                else:
-                    log.msg("Indexing resource: %s" % (name,))
-                    self.addResource(name, calendar, True)
+                calendar = Component.fromStream(stream)
+                calendar.validateForCalDAV()
+            except ValueError:
+                log.err("Non-calendar resource: %s" % (name,))
+            else:
+                #log.msg("Indexing resource: %s" % (name,))
+                self.addResource(name, calendar, True, reCreate=True)
             finally:
                 stream.close()
-        
+
         # Do commit outside of the loop for better performance
-        self._db_commit()
+        if do_commit:
+            self._db_commit()
 
 class IndexSchedule (CalendarIndex):
     """
@@ -610,30 +816,30 @@ class IndexSchedule (CalendarIndex):
         @param uid: the UID to reserve
         @raise ReservationError: if C{uid} is already reserved
         """
-        
+
         # iTIP does not require unique UIDs
-        pass
-    
+        return succeed(None)
+
     def unreserveUID(self, uid): #@UnusedVariable
         """
         Unreserve a UID for this index's resource.
         @param uid: the UID to reserve
         @raise ReservationError: if C{uid} is not reserved
         """
-        
+
         # iTIP does not require unique UIDs
-        pass
-    
+        return succeed(None)
+
     def isReservedUID(self, uid): #@UnusedVariable
         """
         Check to see whether a UID is reserved.
         @param uid: the UID to check
         @return: True if C{uid} is reserved, False otherwise.
         """
-        
+
         # iTIP does not require unique UIDs
-        return False
-        
+        return succeed(False)
+
     def isAllowedUID(self, uid, *names): #@UnusedVariable
         """
         Checks to see whether to allow an operation with adds the the specified
@@ -646,30 +852,30 @@ class IndexSchedule (CalendarIndex):
         @return: True if the UID is not in the index and is not reserved,
             False otherwise.
         """
-        
+
         # iTIP does not require unique UIDs
         return True
- 
+
     def _db_type(self):
         """
         @return: the collection type assigned to this index.
         """
         return collection_types["iTIP"]
-        
+
     def _db_init_data_tables(self, q):
         """
         Initialise the underlying database tables.
         @param q:           a database cursor to use.
         """
-        
+
         # Create database where the RESOURCE table has a UID column that is not unique.
         self._db_init_data_tables_base(q, False)
 
-    def _db_recreate(self):
+    def _db_recreate(self, do_commit=True):
         """
         Re-create the database tables from existing calendar data.
         """
-        
+
         #
         # Populate the DB with data from already existing resources.
         # This allows for index recovery if the DB file gets
@@ -686,16 +892,19 @@ class IndexSchedule (CalendarIndex):
                 log.err("Unable to open resource %s: %s" % (name, e))
                 continue
 
+            # FIXME: This is blocking I/O
             try:
-                # FIXME: This is blocking I/O
-                try:
-                    calendar = Component.fromStream(stream)
-                    calendar.validCalendarForCalDAV()
-                    calendar.validateComponentsForCalDAV(True)
-                except ValueError:
-                    log.err("Non-calendar resource: %s" % (name,))
-                else:
-                    log.msg("Indexing resource: %s" % (name,))
-                    self.addResource(name, calendar)
+                calendar = Component.fromStream(stream)
+                calendar.validCalendarForCalDAV()
+                calendar.validateComponentsForCalDAV(True)
+            except ValueError:
+                log.err("Non-calendar resource: %s" % (name,))
+            else:
+                #log.msg("Indexing resource: %s" % (name,))
+                self.addResource(name, calendar, True, reCreate=True)
             finally:
                 stream.close()
+
+        # Do commit outside of the loop for better performance
+        if do_commit:
+            self._db_commit()
