@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2006-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2006-2009 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Cyrus Daboo, cdaboo@apple.com
 ##
 
 """
@@ -26,7 +24,7 @@ from twistedcaldav.dateops import normalizeForIndex, compareDateTime, difference
 
 from vobject.icalendar import utc
 
-# The maximum number of instances we will ezpand out to.
+# The maximum number of instances we will expand out to.
 # Raise a TooManyInstancesError exception if we exceed this.
 max_allowed_instances = 1000
 
@@ -38,6 +36,14 @@ class TooManyInstancesError(Exception):
     def __repr__(self):
         return "<%s max:%s>" % (self.__class__.__name__, self.max_allowed)
 
+class InvalidOverriddenInstanceError(Exception):
+    def __init__(self, rid):
+        Exception.__init__(self)
+        self.rid = rid
+
+    def __repr__(self):
+        return "<%s invalid:%s>" % (self.__class__.__name__, self.rid)
+
 class Instance(object):
     
     __slots__ = ["component", "start", "end", "rid", "overridden", "future"]
@@ -90,11 +96,12 @@ class Instance(object):
     
 class InstanceList(object):
     
-    __slots__ = ["instances", "limit"]
+    __slots__ = ["instances", "limit", "ignoreInvalidInstances",]
     
-    def __init__(self):
+    def __init__(self, ignoreInvalidInstances=False):
         self.instances = {}
         self.limit = None
+        self.ignoreInvalidInstances = ignoreInvalidInstances
         
     def __iter__(self):
         # Return keys in sorted order via iterator
@@ -117,6 +124,7 @@ class InstanceList(object):
         """
         
         # Look at each component type
+        got_master = False
         overrides = []
         for component in componentSet:
             if component.name() == "VEVENT":
@@ -124,11 +132,13 @@ class InstanceList(object):
                     overrides.append(component)
                 else:
                     self._addMasterEventComponent(component, limit)
+                    got_master = True
             elif component.name() == "VTODO":
                 if component.hasProperty("RECURRENCE-ID"):
                     overrides.append(component)
                 else:
                     self._addMasterToDoComponent(component, limit)
+                    got_master = True
             elif component.name() == "VJOURNAL":
                 #TODO: VJOURNAL
                 raise NotImplementedError("VJOURNAL recurrence expansion not supported yet")
@@ -142,18 +152,19 @@ class InstanceList(object):
                 else:
                     # AVAILABLE components are just like VEVENT components
                     self._addMasterEventComponent(component, limit)
+                    got_master = True
             
         for component in overrides:
             if component.name() == "VEVENT":
-                self._addOverrideEventComponent(component)
+                self._addOverrideEventComponent(component, limit, got_master)
             elif component.name() == "VTODO":
-                self._addOverrideToDoComponent(component)
+                self._addOverrideToDoComponent(component, limit, got_master)
             elif component.name() == "VJOURNAL":
                 #TODO: VJOURNAL
                 raise NotImplementedError("VJOURNAL recurrence expansion not supported yet")
             elif component.name() == "AVAILABLE":
                 # AVAILABLE components are just like VEVENT components
-                self._addOverrideEventComponent(component)
+                self._addOverrideEventComponent(component, limit, got_master)
 
     def addInstance(self, instance):
         """
@@ -193,12 +204,12 @@ class InstanceList(object):
 
         self._addMasterComponent(component, limit, start, end, duration)
 
-    def _addOverrideEventComponent(self, component):
+    def _addOverrideEventComponent(self, component, limit, got_master):
         """
         Add the specified overridden VEVENT Component to the instance list, replacing 
         the one generated by the master component.
         @param component: the overridden Component.
-        @param limit: the end datetime.datetime for expansion.
+        @param got_master: whether a master component has already been expanded.
         """
         
         #TODO: This does not take into account THISANDPRIOR - only THISANDFUTURE
@@ -220,7 +231,7 @@ class InstanceList(object):
         else:
             duration = differenceDateTime(start, end)
 
-        self._addOverrideComponent(component, start, end)
+        self._addOverrideComponent(component, limit, start, end, got_master)
 
     def _addMasterToDoComponent(self, component, limit):
         """
@@ -243,12 +254,12 @@ class InstanceList(object):
 
         self._addMasterComponent(component, limit, start, due, duration)
 
-    def _addOverrideToDoComponent(self, component):
+    def _addOverrideToDoComponent(self, component, limit, got_master):
         """
         Add the specified overridden VTODO Component to the instance list, replacing 
         the one generated by the master component.
         @param component: the overridden Component.
-        @param limit: the end datetime.datetime for expansion.
+        @param got_master: whether a master component has already been expanded.
         """
         
         #TODO: This does not take into account THISANDPRIOR - only THISANDFUTURE
@@ -264,7 +275,7 @@ class InstanceList(object):
         elif due is None:
             due = start
 
-        self._addOverrideComponent(component, start, due)
+        self._addOverrideComponent(component, limit, start, due, got_master)
 
     def _addMasterComponent(self, component, limit, start, end, duration):
         # Always add first instance if included in range.
@@ -272,7 +283,14 @@ class InstanceList(object):
             # dateutils does not do all-day - so convert to datetime.datetime
             start = normalizeForIndex(start)
             end = normalizeForIndex(end)
-            self.addInstance(Instance(component, start, end))
+            
+            # Do not add if in EXDATEs
+            exdates = []
+            for prop in component.properties("EXDATE"):
+                exdates.extend(prop.value())
+            exdates = [normalizeForIndex(exdate) for exdate in exdates]
+            if start not in exdates:
+                self.addInstance(Instance(component, start, end))
         else:
             self.limit = limit
         
@@ -289,16 +307,29 @@ class InstanceList(object):
                 endDate = normalizeForIndex(endDate)
                 self.addInstance(Instance(component, startDate, endDate))
     
-    def _addOverrideComponent(self, component, start, end):
+    def _addOverrideComponent(self, component, limit, start, end, got_master):
 
         # Get the recurrence override info
         rid = component.getRecurrenceIDUTC()
         range = component.getRange()
         
-        # Now add this instance, effectively overridding the one with the matching R-ID
+        # Now add this instance, effectively overriding the one with the matching R-ID
         start = normalizeForIndex(start)
         end = normalizeForIndex(end)
         rid = normalizeForIndex(rid)
+
+        # Make sure start is within the limit
+        if compareDateTime(start, limit) > 0 and compareDateTime(rid, limit) > 0:
+            return
+
+        # Make sure override RECURRENCE-ID is a valid instance of the master
+        if got_master:
+            if str(rid) not in self.instances and compareDateTime(rid, limit) <= 0:
+                if self.ignoreInvalidInstances:
+                    return
+                else:
+                    raise InvalidOverriddenInstanceError(str(rid))
+        
         self.addInstance(Instance(component, start, end, rid, True, range))
         
         # Handle THISANDFUTURE if present
@@ -317,7 +348,7 @@ class InstanceList(object):
             for key in sorted(x for x in self.instances.keys() if x > str(rid)):
                 oldinstance = self.instances[key]
                 
-                # Do not override instance that is alreday overridden
+                # Do not override instance that is already overridden
                 if oldinstance.overridden:
                     continue
                 
@@ -367,7 +398,7 @@ class InstanceList(object):
         Add the specified master VAVAILABILITY Component to the instance list, expanding it
         within the supplied time range. VAVAILABILITY components are not recurring, they have an
         optional DTSTART and DTEND/DURATION defining a single time-range which may be bounded
-        depedning on the presence of the properties. If unbounded at one or both ends, we will
+        depending on the presence of the properties. If unbounded at one or both ends, we will
         set the time to 1/1/1900 in the past and 1/1/3000 in the future.
         @param component: the Component to expand
         @param limit: the end datetime.datetime for expansion
