@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Wilfredo Sanchez, wsanchez@apple.com
 ##
 
 """
@@ -29,6 +27,7 @@ __all__ = [
     "get",
     "mkcalendar",
     "mkcol",
+    "propfind",
     "put",
     "report",
     "report_calquery",
