@@ -1,5 +1,5 @@
 ##
-# Copyright (c) 2006-2007 Apple Inc. All rights reserved.
+# Copyright (c) 2006-2008 Apple Inc. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,8 +12,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Cyrus Daboo, cdaboo@apple.com
 ##
 
 __all__ = [
@@ -22,14 +20,25 @@ __all__ = [
     "allPropertiesForResource",
     "propertyNamesForResource",
     "propertyListForResource",
+    "validPropertyListCalendarDataTypeVersion",
     "generateFreeBusyInfo",
     "processEventFreeBusy",
     "processFreeBusyFreeBusy",
+    "processAvailabilityFreeBusy",
     "buildFreeBusyResult",
 ]
 
-from twisted.internet.defer import deferredGenerator, waitForDeferred
-from twisted.python import log
+import datetime
+import time
+
+try:
+    from hashlib import md5
+except ImportError:
+    from md5 import new as md5
+
+from vobject.icalendar import utc
+
+from twisted.internet.defer import inlineCallbacks, returnValue
 from twisted.python.failure import Failure
 from twisted.web2 import responsecode
 from twisted.web2.dav import davxml
@@ -47,13 +56,12 @@ from twistedcaldav.customxml import TwistedCalendarAccessProperty
 from twistedcaldav.dateops import clipPeriod, normalizePeriodList, timeRangesOverlap
 from twistedcaldav.ical import Component, Property, iCalendarProductID
 from twistedcaldav.instance import InstanceList
+from twistedcaldav.index import IndexedSearchException
+from twistedcaldav.log import Logger
 
-from vobject.icalendar import utc
-
-import datetime
-import md5
-import time
+log = Logger()
 
+@inlineCallbacks
 def applyToCalendarCollections(resource, request, request_uri, depth, apply, privileges):
     """
     Run an operation on all calendar collections, starting at the specified
@@ -70,13 +78,11 @@ def applyToCalendarCollections(resource, request, request_uri, depth, apply, pri
     """
 
     # First check the privilege on this resource
-    try:
-        d = waitForDeferred(resource.checkPrivileges(request, privileges))
-        yield d
-        d.getResult()
-    except AccessDeniedError:
-        yield None
-        return
+    if privileges:
+        try:
+            yield resource.checkPrivileges(request, privileges)
+        except AccessDeniedError:
+            return
 
     # When scanning we only go down as far as a calendar collection - not into one
     if resource.isPseudoCalendarCollection():
@@ -85,25 +91,21 @@ def applyToCalendarCollections(resource, request, request_uri, depth, apply, pri
         resources = [(resource, request_uri)]
     else:
         resources = []
-        d = waitForDeferred(resource.findCalendarCollections(depth, request, lambda x, y: resources.append((x, y)), privileges = privileges))
-        yield d
-        d.getResult()
+        yield resource.findCalendarCollections(depth, request, lambda x, y: resources.append((x, y)), privileges = privileges)
          
     for calresource, uri in resources:
-        d = waitForDeferred(apply(calresource, uri))
-        yield d
-        d.getResult()
+        result = (yield apply(calresource, uri))
+        if not result:
+            break
 
-applyToCalendarCollections = deferredGenerator(applyToCalendarCollections)
-
-def responseForHref(request, responses, href, resource, calendar, propertiesForResource, propertyreq, isowner=True):
+def responseForHref(request, responses, href, resource, calendar, timezone, propertiesForResource, propertyreq, isowner=True):
     """
     Create an appropriate property status response for the given resource.
 
     @param request: the L{IRequest} for the current request.
     @param responses: the list of responses to append the result of this method to.
-    @param href: the L{HRef} element of the resource being targetted.
-    @param resource: the L{CalDAVFile} for the targetted resource.
+    @param href: the L{HRef} element of the resource being targeted.
+    @param resource: the L{CalDAVFile} for the targeted resource.
     @param calendar: the L{Component} for the calendar for the resource. This may be None
         if the calendar has not already been read in, in which case the resource
         will be used to get the calendar if needed.
@@ -114,53 +116,56 @@ def responseForHref(request, responses, href, resource, calendar, propertiesForR
     """
 
     def _defer(properties_by_status):
+        propstats = []
+
         for status in properties_by_status:
             properties = properties_by_status[status]
             if properties:
-                responses.append(
-                    davxml.PropertyStatusResponse(
-                        href,
-                        davxml.PropertyStatus(
-                            davxml.PropertyContainer(*properties),
-                            davxml.Status.fromResponseCode(status)
-                        )
-                    )
-                )
+                xml_status = davxml.Status.fromResponseCode(status)
+                xml_container = davxml.PropertyContainer(*properties)
+                xml_propstat = davxml.PropertyStatus(xml_container, xml_status)
 
-    d = propertiesForResource(request, propertyreq, resource, calendar, isowner)
+                propstats.append(xml_propstat)
+
+        if propstats:
+            responses.append(davxml.PropertyStatusResponse(href, *propstats))
+
+    d = propertiesForResource(request, propertyreq, resource, calendar, timezone, isowner)
     d.addCallback(_defer)
     return d
 
-def allPropertiesForResource(request, prop, resource, calendar=None, isowner=True): #@UnusedVariable
+def allPropertiesForResource(request, prop, resource, calendar=None, timezone=None, isowner=True):
     """
     Return all (non-hidden) properties for the specified resource.
     @param request: the L{IRequest} for the current request.
     @param prop: the L{PropertyContainer} element for the properties of interest.
-    @param resource: the L{CalDAVFile} for the targetted resource.
+    @param resource: the L{CalDAVFile} for the targeted resource.
     @param calendar: the L{Component} for the calendar for the resource. This may be None
         if the calendar has not already been read in, in which case the resource
         will be used to get the calendar if needed.
+    @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
     @param isowner: C{True} if the authorized principal making the request is the DAV:owner,
         C{False} otherwise.
     @return: a map of OK and NOT FOUND property values.
     """
 
     def _defer(props):
-        return _namedPropertiesForResource(request, props, resource, calendar, isowner)
+        return _namedPropertiesForResource(request, props, resource, calendar, timezone, isowner)
 
     d = resource.listAllprop(request)
     d.addCallback(_defer)
     return d
 
-def propertyNamesForResource(request, prop, resource, calendar=None, isowner=True): #@UnusedVariable
+def propertyNamesForResource(request, prop, resource, calendar=None, timezone=None, isowner=True): #@UnusedVariable
     """
     Return property names for all properties on the specified resource.
     @param request: the L{IRequest} for the current request.
     @param prop: the L{PropertyContainer} element for the properties of interest.
-    @param resource: the L{CalDAVFile} for the targetted resource.
+    @param resource: the L{CalDAVFile} for the targeted resource.
     @param calendar: the L{Component} for the calendar for the resource. This may be None
         if the calendar has not already been read in, in which case the resource
         will be used to get the calendar if needed.
+    @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
     @param isowner: C{True} if the authorized principal making the request is the DAV:owner,
         C{False} otherwise.
     @return: a map of OK and NOT FOUND property values.
@@ -176,21 +181,22 @@ def propertyNamesForResource(request, prop, resource, calendar=None, isowner=Tru
     d.addCallback(_defer)
     return d
 
-def propertyListForResource(request, prop, resource, calendar=None, isowner=True):
+def propertyListForResource(request, prop, resource, calendar=None, timezone=None, isowner=True):
     """
     Return the specified properties on the specified resource.
     @param request: the L{IRequest} for the current request.
     @param prop: the L{PropertyContainer} element for the properties of interest.
-    @param resource: the L{CalDAVFile} for the targetted resource.
+    @param resource: the L{CalDAVFile} for the targeted resource.
     @param calendar: the L{Component} for the calendar for the resource. This may be None
         if the calendar has not already been read in, in which case the resource
         will be used to get the calendar if needed.
+    @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
     @param isowner: C{True} if the authorized principal making the request is the DAV:owner,
         C{False} otherwise.
     @return: a map of OK and NOT FOUND property values.
     """
     
-    return _namedPropertiesForResource(request, prop.children, resource, calendar, isowner)
+    return _namedPropertiesForResource(request, prop.children, resource, calendar, timezone, isowner)
 
 def validPropertyListCalendarDataTypeVersion(prop):
     """
@@ -215,15 +221,17 @@ def validPropertyListCalendarDataTypeVersion(prop):
 
     return result, message, generate_calendar_data
 
-def _namedPropertiesForResource(request, props, resource, calendar=None, isowner=True):
+@inlineCallbacks
+def _namedPropertiesForResource(request, props, resource, calendar=None, timezone=None, isowner=True):
     """
     Return the specified properties on the specified resource.
     @param request: the L{IRequest} for the current request.
     @param props: a list of property elements or qname tuples for the properties of interest.
-    @param resource: the L{CalDAVFile} for the targetted resource.
+    @param resource: the L{CalDAVFile} for the targeted resource.
     @param calendar: the L{Component} for the calendar for the resource. This may be None
         if the calendar has not already been read in, in which case the resource
         will be used to get the calendar if needed.
+    @param timezone: the L{Component} the VTIMEZONE to use for floating/all-day.
     @param isowner: C{True} if the authorized principal making the request is the DAV:owner,
         C{False} otherwise.
     @return: a map of OK and NOT FOUND property values.
@@ -245,9 +253,9 @@ def _namedPropertiesForResource(request, props, resource, calendar=None, isowner
                 access = None
 
             if calendar:
-                propvalue = property.elementFromCalendarWithAccessRestrictions(calendar, access)
+                propvalue = property.elementFromCalendarWithAccessRestrictions(calendar, access, timezone)
             else:
-                propvalue = property.elementFromResourceWithAccessRestrictions(resource, access)
+                propvalue = property.elementFromResourceWithAccessRestrictions(resource, access, timezone)
             if propvalue is None:
                 raise ValueError("Invalid CalDAV:calendar-data for request: %r" % (property,))
             properties_by_status[responsecode.OK].append(propvalue)
@@ -258,15 +266,11 @@ def _namedPropertiesForResource(request, props, resource, calendar=None, isowner
         else:
             qname = property
     
-        props = waitForDeferred(resource.listProperties(request))
-        yield props
-        props = props.getResult()
+        has = (yield resource.hasProperty(property, request))
 
-        if qname in props:
+        if has:
             try:
-                prop = waitForDeferred(resource.readProperty(qname, request))
-                yield prop
-                prop = prop.getResult()
+                prop = (yield resource.readProperty(qname, request))
                 properties_by_status[responsecode.OK].append(prop)
             except HTTPError:
                 f = Failure()
@@ -279,12 +283,12 @@ def _namedPropertiesForResource(request, props, resource, calendar=None, isowner
         else:
             properties_by_status[responsecode.NOT_FOUND].append(propertyName(qname))
     
-    yield properties_by_status
+    returnValue(properties_by_status)
 
-_namedPropertiesForResource = deferredGenerator(_namedPropertiesForResource)
-    
+@inlineCallbacks
 def generateFreeBusyInfo(request, calresource, fbinfo, timerange, matchtotal,
-                         excludeuid=None, organizer=None, same_calendar_user=False):
+                         excludeuid=None, organizer=None, same_calendar_user=False,
+                         servertoserver=False):
     """
     Run a free busy report on the specified calendar collection
     accumulating the free busy info for later processing.
@@ -295,20 +299,24 @@ def generateFreeBusyInfo(request, calresource, fbinfo, timerange, matchtotal,
     @param matchtotal:  the running total for the number of matches.
     @param excludeuid:  a C{str} containing a UID value to exclude any components with that
         UID from contributing to free-busy.
-    @param organizer:   a C{str} containing the value of the ORGANIZER proeprty in the VFREEBUSY request.
+    @param organizer:   a C{str} containing the value of the ORGANIZER property in the VFREEBUSY request.
         This is used in conjunction with the UID value to process exclusions.
-    @param same_calendar_user:   a C{bool} indicating whether the calendar user requesting tyhe free-busy information
+    @param same_calendar_user:   a C{bool} indicating whether the calendar user requesting the free-busy information
         is the same as the calendar user being targeted.
+    @param servertoserver:       a C{bool} indicating whether we are doing a local or remote lookup request.
     """
     
     # First check the privilege on this collection
-    try:
-        d = waitForDeferred(calresource.checkPrivileges(request, (caldavxml.ReadFreeBusy(),)))
-        yield d
-        d.getResult()
-    except AccessDeniedError:
-        yield matchtotal
-        return
+    # TODO: for server-to-server we bypass this right now as we have no way to authorize external users.
+    if not servertoserver:
+        try:
+            yield calresource.checkPrivileges(request, (caldavxml.ReadFreeBusy(),))
+        except AccessDeniedError:
+            returnValue(matchtotal)
+
+    # May need organizer principal
+    organizer_principal = calresource.principalForCalendarUserAddress(organizer) if organizer else None
+    organizer_uid = organizer_principal.principalUID() if organizer_principal else ""
 
     #
     # What we do is a fake calendar-query for VEVENT/VFREEBUSYs in the specified time-range.
@@ -329,42 +337,39 @@ def generateFreeBusyInfo(request, calresource, fbinfo, timerange, matchtotal,
 
     # Get the timezone property from the collection, and store in the query filter
     # for use during the query itself.
-    has_prop = waitForDeferred(calresource.hasProperty((caldav_namespace, "calendar-timezone"), request))
-    yield has_prop
-    has_prop = has_prop.getResult()
+    has_prop = (yield calresource.hasProperty((caldav_namespace, "calendar-timezone"), request))
     if has_prop:
-        tz = waitForDeferred(calresource.readProperty((caldav_namespace, "calendar-timezone"), request))
-        yield tz
-        tz = tz.getResult()
+        tz = (yield calresource.readProperty((caldav_namespace, "calendar-timezone"), request))
     else:
         tz = None
     tzinfo = filter.settimezone(tz)
 
-    # Do some optimisation of access control calculation by determining any inherited ACLs outside of
+    # Do some optimization of access control calculation by determining any inherited ACLs outside of
     # the child resource loop and supply those to the checkPrivileges on each child.
-    filteredaces = waitForDeferred(calresource.inheritedACEsforChildren(request))
-    yield filteredaces
-    filteredaces = filteredaces.getResult()
+    filteredaces = (yield calresource.inheritedACEsforChildren(request))
+
+    try:
+        resources = calresource.index().indexedSearch(filter)
+    except IndexedSearchException:
+        resources = calresource.index().bruteForceSearch()
+
+    for name, uid, type in resources: #@UnusedVariable
 
-    for name, uid, type in calresource.index().search(filter): #@UnusedVariable
-        
         # Check privileges - must have at least CalDAV:read-free-busy
-        child = waitForDeferred(request.locateChildResource(calresource, name))
-        yield child
-        child = child.getResult()
+        child = (yield request.locateChildResource(calresource, name))
 
-        try:
-            d = waitForDeferred(child.checkPrivileges(request, (caldavxml.ReadFreeBusy(),), inherited_aces=filteredaces))
-            yield d
-            d.getResult()
-        except AccessDeniedError:
-            continue
+        # TODO: for server-to-server we bypass this right now as we have no way to authorize external users.
+        if not servertoserver:
+            try:
+                yield child.checkPrivileges(request, (caldavxml.ReadFreeBusy(),), inherited_aces=filteredaces)
+            except AccessDeniedError:
+                continue
 
         calendar = calresource.iCalendar(name)
         
         # The calendar may come back as None if the resource is being changed, or was deleted
-        # between our initial index query and getting here. For now we will ignore this errror, but in
-        # the longer term we need to simplement some form of locking, perhaps.
+        # between our initial index query and getting here. For now we will ignore this error, but in
+        # the longer term we need to implement some form of locking, perhaps.
         if calendar is None:
             log.err("Calendar %s is missing from calendar collection %r" % (name, calresource))
             continue
@@ -373,18 +378,22 @@ def generateFreeBusyInfo(request, calresource, fbinfo, timerange, matchtotal,
         if excludeuid:
             # See if we have a UID match
             if (excludeuid == uid):
+                test_organizer = calendar.getOrganizer()
+                test_principal = calresource.principalForCalendarUserAddress(test_organizer) if test_organizer else None
+                test_uid = test_principal.principalUID() if test_principal else ""
+
                 # Check that ORGANIZER's match (security requirement)
-                if (organizer is None) or (organizer == calendar.getOrganizer()):
+                if (organizer is None) or (organizer_uid == test_uid):
                     continue
                 # Check for no ORGANIZER and check by same calendar user
-                elif (calendar.getOrganizer() is None) and same_calendar_user:
+                elif (test_organizer is None) and same_calendar_user:
                     continue
 
         if filter.match(calendar, None):
             # Check size of results is within limit
             matchtotal += 1
             if matchtotal > max_number_of_matches:
-                raise NumberOfMatchesWithinLimits
+                raise NumberOfMatchesWithinLimits(max_number_of_matches)
 
             if calendar.mainType() == "VEVENT":
                 processEventFreeBusy(calendar, fbinfo, timerange, tzinfo)
@@ -395,9 +404,7 @@ def generateFreeBusyInfo(request, calresource, fbinfo, timerange, matchtotal,
             else:
                 assert "Free-busy query returned unwanted component: %s in %r", (name, calresource,)
     
-    yield matchtotal
-
-generateFreeBusyInfo = deferredGenerator(generateFreeBusyInfo)
+    returnValue(matchtotal)
 
 def processEventFreeBusy(calendar, fbinfo, timerange, tzinfo):
     """
@@ -409,7 +416,7 @@ def processEventFreeBusy(calendar, fbinfo, timerange, tzinfo):
     """
     
     # Expand out the set of instances for the event with in the required range
-    instances = calendar.expandTimeRanges(timerange.end)
+    instances = calendar.expandTimeRanges(timerange.end, ignoreInvalidInstances=True)
     
     # Can only do timed events
     for key in instances:
@@ -488,7 +495,7 @@ def processFreeBusyFreeBusy(calendar, fbinfo, timerange):
             if fbtype == "FREE":
                 continue
             
-            # Look at each period in the propert
+            # Look at each period in the property
             assert isinstance(fb.value(), list), "FREEBUSY property does not contain a list of values: %r" % (fb,)
             for period in fb.value():
                 # Clip period for this instance
@@ -557,9 +564,9 @@ def processAvailablePeriods(calendar, timerange):
             uid = component.propertyValue("UID")
             uidmap.setdefault(uid, []).append(component)
             
-    # Then we expand each uid set seperately
+    # Then we expand each uid set separately
     for componentSet in uidmap.itervalues():
-        instances = InstanceList()
+        instances = InstanceList(ignoreInvalidInstances=True)
         instances.expandTimeRanges(componentSet, timerange.end)
         
         # Now convert instances into period list
@@ -628,7 +635,7 @@ def buildFreeBusyResult(fbinfo, timerange, organizer=None, attendee=None, uid=No
     if uid is not None:
         fb.addProperty(Property("UID", uid))
     else:
-        uid = md5.new(str(fbcalendar) + str(time.time())).hexdigest()
+        uid = md5(str(fbcalendar) + str(time.time())).hexdigest()
         fb.addProperty(Property("UID", uid))
 
     return fbcalendar
