@@ -12,12 +12,8 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-#
-# DRI: Cyrus Daboo, cdaboo@apple.com
 ##
 
-__all__ = [
-    "expression",
-    "sqlgenerator",
-    "caldavquery",
-]
+"""
+CalDAV queries.
+"""
